/* ------------------ **
**      Includes      **
** ------------------ */

#include "Yap.h"
#ifdef YAPOR
#include "or.macros.h"
#ifdef TABLING
#include "tab.macros.h"
#endif /* TABLING */



/* ---------------------- **
**      Local macros      **
** ---------------------- */

STD_PROTO(static inline void PUT_NO_WORK_IN_UP_CPS, (void));
STD_PROTO(static inline void PUT_ME_IDLE, (int));
STD_PROTO(static inline void PUT_ME_BUSY, (int));
STD_PROTO(static inline void PUT_ME_IN_ROOT_CP, (int));
STD_PROTO(static inline void PUT_ME_OUT_ROOT_CP, (int));

/*
static inline 
void MOVE_UP_ONE_NODE(choiceptr nearest_livenode) {
  LOCK_OR_FRAME(LOCAL_top_or_fr);
  BITMAP_delete(LOCAL_top_or_fr->members, worker_id);
  if (LOCAL_top_or_fr->pending_cut_cp != NULL &&
      GLOBAL_branch_ltt(worker_id, LOCAL_top_or_fr->depth) >= LOCAL_top_or_fr->pending_cut_ltt &&
      no_workers_left_pending_cut()) {
    CUT_continue_pending_cut();
  } else if (BITMAP_empty(LOCAL_top_or_fr->members)) {
    sol_fr_ptr solution;
    solution = LOCAL_top_or_fr->solutions;
    CUT_put_solutions_together(solution);
    FREE_OR_FRAME(LOCAL_top_or_fr);
    SCH_update_local_or_tops();
    if (solution != NULL) {
      choiceptr leftmost_cp;
      or_fr_ptr leftmost_or_fr;
      leftmost_cp = CUT_leftmost_until();
      leftmost_or_fr = leftmost_cp->cp_or_fr;
      LOCK_OR_FRAME(leftmost_or_fr);
      if (LOCAL_cut_cp >= LOCAL_top_cp) {
        UNLOCK_OR_FRAME(leftmost_or_fr);
        CUT_free_pending_solutions(solution);
        CUT_move_up_to_cut_cp();
      } else {
        CUT_put_inherted_solutions(solution, leftmost_or_fr);
        UNLOCK_OR_FRAME(leftmost_or_fr);
      }
    }
    CUT_clear_cut_cp();
  } else {
    if (nearest_livenode)
      LOCAL_top_or_fr->nearest_livenode = nearest_livenode;
    UNLOCK_OR_FRAME(LOCAL_top_or_fr);
    SCH_update_local_or_tops();
    CUT_clear_cut_cp();
  }
  return;
}
*/

static inline
void PUT_NO_WORK_IN_UP_CPS(void) {
  or_fr_ptr aux1, aux2;
  aux1 = LOCAL_top_or_fr;
  aux2 = OrFr_nearest_livenode(aux1);
  while (aux2) {
    OrFr_nearest_livenode(aux1) = NULL;
    aux1 = aux2;
    aux2 = OrFr_nearest_livenode(aux1);
  }
  return;
}


static inline
void PUT_ME_IDLE(int worker_num) {
  LOCK(GLOBAL_LOCKS_bm_idle_workers);
  BITMAP_insert(GLOBAL_bm_idle_workers, worker_num);
  UNLOCK(GLOBAL_LOCKS_bm_idle_workers);
  return;
}


static inline
void PUT_ME_BUSY(int worker_num) {
  LOCK(GLOBAL_LOCKS_bm_idle_workers);
  BITMAP_delete(GLOBAL_bm_idle_workers, worker_num);
  UNLOCK(GLOBAL_LOCKS_bm_idle_workers);
  return;
}


static inline
void PUT_ME_IN_ROOT_CP(int worker_num) {
  LOCK(GLOBAL_LOCKS_bm_root_cp_workers);
  BITMAP_insert(GLOBAL_bm_root_cp_workers, worker_num);
  UNLOCK(GLOBAL_LOCKS_bm_root_cp_workers);
  return;
}


static inline
void PUT_ME_OUT_ROOT_CP(int worker_num) {
  LOCK(GLOBAL_LOCKS_bm_root_cp_workers);
  BITMAP_delete(GLOBAL_bm_root_cp_workers, worker_num);
  UNLOCK(GLOBAL_LOCKS_bm_root_cp_workers);
  return;
}



/* ------------------------------------- **
**      Local functions declaration      **
** ------------------------------------- */

static int get_work_below(void);
static int get_work_above(void);
static int find_a_better_position(void);
static int search_for_hidden_shared_work(bitmap stable_busy);



/* -------------------------- **
**      Global functions      **
** -------------------------- */

/* 
** move up one node --> ??? check for frozen stacks on branch ???
**    frozen stacks on branch --> return TRUE;
**    no frozen stacks on branch --> ??? check for workers in node ???
**       other workers in node --> return TRUE;
**       no more workers in node --> ??? check for unconsumed answers in suspension frames ???
**          unconsumed answers in suspension frame --> return FALSE;
**          no suspension frames with unconsumed answers --> ??? check for hidden workers in node ???
**             hidden workers in node --> return TRUE;
**             no hidden workers in node --> return TRUE;
*/

int move_up_one_node(or_fr_ptr nearest_livenode) {
  LOCK_OR_FRAME(LOCAL_top_or_fr);
  BITMAP_delete(OrFr_members(LOCAL_top_or_fr), worker_id);

  /* continue pending cuts ? */
  if (OrFr_pend_cut_cp(LOCAL_top_or_fr) != NULL
      && GLOBAL_branch_ltt(worker_id, OrFr_depth(LOCAL_top_or_fr)) >= OrFr_pend_cut_ltt(LOCAL_top_or_fr)
      && no_workers_left_pending_cut()) {
    CUT_continue_pending_cut();
    return TRUE;
  }

  /* check for frozen stacks on branch */
#ifdef OPTYAP_ERRORS
  if (B_FZ != DepFr_b_fz(LOCAL_top_dep_fr))
    OPTYAP_ERROR_MESSAGE("B_FZ != DepFr_b_fz(LOCAL_top_dep_fr) (move_up_one_node)");
#endif /* OPTYAP_ERRORS */
#ifdef TABLING
  if (EQUAL_OR_YOUNGER_CP(B_FZ, LOCAL_top_cp)) {
    /* frozen stacks on branch */
    BITMAP_insert(OrFr_hidden_members(LOCAL_top_or_fr), worker_id);
    if (BITMAP_empty(OrFr_members(LOCAL_top_or_fr))) {
      /* the last worker leaving the node is the one that 
         collects suspension frames with unconsumed answers */
      collect_suspension_frames_with_unconsumed_answers();
    }
    if (nearest_livenode)
      OrFr_nearest_livenode(LOCAL_top_or_fr) = nearest_livenode;
    UNLOCK_OR_FRAME(LOCAL_top_or_fr);
    SCH_update_local_or_tops();
    CUT_clear_cut_cp();
    return TRUE;
  }
#endif /* TABLING */

  /* no frozen stacks on branch --> check for workers in node */
#ifdef OPTYAP_ERRORS
  if (LOCAL_top_cp != LOCAL_top_cp_on_stack)
    OPTYAP_ERROR_MESSAGE("LOCAL_top_cp != LOCAL_top_cp_on_stack (move_up_one_node)");
#endif /* OPTYAP_ERRORS */
  if (BITMAP_empty(OrFr_members(LOCAL_top_or_fr))) {
    /* no frozen stacks on branch and no more workers in node */
    sol_fr_ptr solution;
#ifdef TABLING
    susp_fr_ptr *susp_ptr, collect_susp_fr;
    susp_ptr = & OrFr_suspensions(LOCAL_top_or_fr);
    collect_susp_fr = *susp_ptr;
    /* check for unconsumed answers in suspension frames */
    while (collect_susp_fr) {
      if (suspension_frame_with_unconsumed_answers(collect_susp_fr)) {
        /* unconsumed answers in suspension frame */


printf("***** DANGER W%d: unconsumed answers in suspension frame %x na or-frame %x\n", worker_id, collect_susp_fr,LOCAL_top_or_fr);



        *susp_ptr = SuspFr_next(collect_susp_fr);
        BITMAP_insert(OrFr_members(LOCAL_top_or_fr), worker_id);
        UNLOCK_OR_FRAME(LOCAL_top_or_fr);

unbind_variables(TR, LOCAL_top_cp->cp_tr);

        /* resume suspension frame */
        resume_suspension_frame(collect_susp_fr);

/* update TR register */
TR = TR_FZ;
TRAIL_REF(LOCAL_top_cp->cp_tr);

        return FALSE;
      }
      susp_ptr = & SuspFr_next(*susp_ptr);
      collect_susp_fr = *susp_ptr;
    }
    /* no suspension frames with unconsumed answers --> check for hidden workers in node */
    if (! BITMAP_empty(OrFr_hidden_members(LOCAL_top_or_fr))) {
      /* no suspension frames with unconsumed answers, but hidden workers in node */
      if (nearest_livenode)
        OrFr_nearest_livenode(LOCAL_top_or_fr) = nearest_livenode;
      UNLOCK_OR_FRAME(LOCAL_top_or_fr);
      LOCAL_top_cp_on_stack = OrFr_node(OrFr_next_on_stack(LOCAL_top_or_fr));
      SCH_update_local_or_tops();
      CUT_clear_cut_cp();
      return TRUE;
    }
    /* no suspension frames with unconsumed answers and no hidden workers in node */
    if ((collect_susp_fr = OrFr_suspensions(LOCAL_top_or_fr)) != NULL)
      complete_suspension_frames(collect_susp_fr);
#endif /* TABLING */
    solution = OrFr_solutions(LOCAL_top_or_fr);
#ifdef TABLING
    LOCAL_top_cp_on_stack = OrFr_node(OrFr_next_on_stack(LOCAL_top_or_fr));
#endif /* TABLING */
    FREE_OR_FRAME(LOCAL_top_or_fr);
    SCH_update_local_or_tops();
    if (solution) {
      choiceptr leftmost_cp;
      or_fr_ptr leftmost_or_fr;
      CUT_put_solutions_together(solution);
      leftmost_cp = CUT_leftmost_until();
      leftmost_or_fr = leftmost_cp->cp_or_fr;
      LOCK_OR_FRAME(leftmost_or_fr);
      if (LOCAL_cut_cp != NULL && YOUNGER_CP(LOCAL_top_cp, LOCAL_cut_cp)) {
        UNLOCK_OR_FRAME(leftmost_or_fr);
        CUT_free_pending_solutions(solution);
      } else {
        CUT_put_inherted_solutions(solution, leftmost_or_fr);
        UNLOCK_OR_FRAME(leftmost_or_fr);
      }
    }
    CUT_clear_cut_cp();
    return TRUE;
  }

  /* no frozen stacks on branch, but other workers in node */
  if (nearest_livenode)
    OrFr_nearest_livenode(LOCAL_top_or_fr) = nearest_livenode;
  UNLOCK_OR_FRAME(LOCAL_top_or_fr);
#ifdef TABLING
  LOCAL_top_cp_on_stack = OrFr_node(OrFr_next_on_stack(LOCAL_top_or_fr));
#endif /* TABLING */
  SCH_update_local_or_tops();
  CUT_clear_cut_cp();
  return TRUE;
}


int get_work(void) {
  int counter;
  bitmap stable_busy;
  yamop *alt_with_work;
  or_fr_ptr frame_with_work;
#ifdef TABLING
  choiceptr leader_cp;
#endif /* TABLING */

  while (LOCAL_reply_signal == install) continue;
  /* find nearest node with available work */
  frame_with_work = LOCAL_top_or_fr;
#ifdef TABLING
  leader_cp = DepFr_leader_cp(LOCAL_top_dep_fr);
  if (leader_cp) {
    do {
      frame_with_work = OrFr_nearest_livenode(frame_with_work);
      if (frame_with_work == NULL || EQUAL_OR_YOUNGER_CP(leader_cp, OrFr_node(frame_with_work))) {
        /* there is a leader node before the nearest node with work */
        frame_with_work = leader_cp->cp_or_fr;
        break;
      }
      alt_with_work = OrFr_alternative(frame_with_work);
    } while (alt_with_work == NULL || alt_with_work->YAMOP_SEQ);
  } else
#endif /* TABLING */
  { do {
      frame_with_work = OrFr_nearest_livenode(frame_with_work);
      if (frame_with_work == NULL)
        break;
      alt_with_work = OrFr_alternative(frame_with_work);
    } while (alt_with_work == NULL || alt_with_work->YAMOP_SEQ);
  }
  if (frame_with_work) {
    do {
      if (! move_up_one_node(frame_with_work)) {
        return TRUE;
      }
    } while (LOCAL_top_or_fr != frame_with_work);
    return TRUE;
  }
  PUT_NO_WORK_IN_UP_CPS();

  PUT_ME_OUT_ROOT_CP(worker_id);
  LOCK_WORKER(worker_id);
  PUT_ME_IDLE(worker_id);
  if (SCH_any_request()) {
    /* if any request to share work, refuse it */
    REMOTE_reply_signal(LOCAL_request) = refused;
    LOCAL_request = MAX_WORKERS;
  }
  UNLOCK_WORKER(worker_id);

  counter = 1;
  BITMAP_difference(stable_busy, OrFr_members(LOCAL_top_or_fr), GLOBAL_bm_idle_workers);
  while (1) {
    while (BITMAP_subset(GLOBAL_bm_idle_workers, OrFr_members(LOCAL_top_or_fr)) &&
           LOCAL_top_cp != GLOBAL_root_cp) {
      /* no busy workers here and below */
      if (OrFr_alternative(LOCAL_top_or_fr) != NULL) {
        /* This is a sequential choicepoint with unexploited alternatives */
        PUT_ME_BUSY(worker_id);
        return TRUE;   
      }
      if (! move_up_one_node(NULL)) {
        PUT_ME_BUSY(worker_id);
        return TRUE;
      }
    }
    if (LOCAL_top_cp == GLOBAL_root_cp) {
      PUT_ME_IN_ROOT_CP(worker_id);
      if (BITMAP_same(GLOBAL_bm_idle_workers, GLOBAL_bm_root_cp_workers) &&
          BITMAP_same(GLOBAL_bm_idle_workers, GLOBAL_bm_present_workers)) {
        /* All workers are idle in root choicepoint. Execution 
           must finish as there is no available computation. */
        return FALSE;
      }
    }
    if (get_work_below()) {
      PUT_ME_BUSY(worker_id);
      return TRUE;
    }
    if (get_work_above()) {
      PUT_ME_BUSY(worker_id);
      return TRUE;
    }
    if (find_a_better_position()) {
      PUT_ME_BUSY(worker_id);
      return TRUE;
    }
    if (counter++ == SCHEDULER_LOOP) {
      if (search_for_hidden_shared_work(stable_busy)) {
        PUT_ME_BUSY(worker_id);
        return TRUE;
      }
      BITMAP_difference(stable_busy, OrFr_members(LOCAL_top_or_fr), GLOBAL_bm_idle_workers);
      counter = 1;
    }
    BITMAP_minus(stable_busy, GLOBAL_bm_idle_workers);
  }
}



/* ------------------------- **
**      Local functions      **
** ------------------------- */

static int get_work_below(void){
  int i, worker_p, big_load;
  bitmap busy_below, idle_below;

  worker_p = -1;
  big_load = 0;
  BITMAP_difference(busy_below, OrFr_members(LOCAL_top_or_fr), GLOBAL_bm_idle_workers);
  BITMAP_difference(idle_below, OrFr_members(LOCAL_top_or_fr), busy_below);
  BITMAP_delete(idle_below, worker_id);
  for (i = 0; i < number_workers; i++) {
    if (BITMAP_member(idle_below ,i) && YOUNGER_CP(REMOTE_top_cp(i), LOCAL_top_cp))
      BITMAP_minus(busy_below, OrFr_members(REMOTE_top_or_fr(i)));
  }
  if (BITMAP_empty(busy_below))
    return FALSE;
  /* choose the worker with highest load */
  for (i = 0 ; i < number_workers; i++) {
    if (BITMAP_member(busy_below ,i) && REMOTE_load(i) > big_load) {
      worker_p = i;
      big_load = REMOTE_load(i);
    }
  }
  if (worker_p == -1) 
    return FALSE;
  return (q_share_work(worker_p));
}


static int get_work_above(void){
  int i, worker_p, big_load;
  bitmap visible_busy_above, visible_idle_above;

  worker_p = -1; 
  big_load = 0;
  BITMAP_difference(visible_busy_above, GLOBAL_bm_present_workers, OrFr_members(LOCAL_top_or_fr));
  BITMAP_minus(visible_busy_above, GLOBAL_bm_invisible_workers);
  BITMAP_copy(visible_idle_above, visible_busy_above); 
  BITMAP_minus(visible_busy_above, GLOBAL_bm_idle_workers);
  BITMAP_and(visible_idle_above, GLOBAL_bm_idle_workers);
  BITMAP_insert(visible_busy_above, worker_id);
  for (i = 0 ; i < number_workers; i++) {
    if (BITMAP_member(visible_idle_above, i))
      BITMAP_minus(visible_busy_above, OrFr_members(REMOTE_top_or_fr(i)));
  }
  if (!BITMAP_member(visible_busy_above, worker_id) || BITMAP_alone(visible_busy_above, worker_id))
    return FALSE;
  BITMAP_delete(visible_busy_above, worker_id);
  /* choose the worker with higher load */
  for (i = 0; i < number_workers; i++) {
    if (BITMAP_member(visible_busy_above ,i) && REMOTE_load(i) > big_load) {
      worker_p = i;
      big_load = REMOTE_load(i);
    }
  }
  if (worker_p == -1)
    return FALSE;
  /* put workers invisibles */
  LOCK(GLOBAL_LOCKS_bm_invisible_workers);
  if (BITMAP_member(GLOBAL_bm_invisible_workers, worker_p)) {
    UNLOCK(GLOBAL_LOCKS_bm_invisible_workers);
    return FALSE;
  }
  BITMAP_insert(GLOBAL_bm_invisible_workers, worker_id);
  BITMAP_insert(GLOBAL_bm_invisible_workers, worker_p);
  UNLOCK(GLOBAL_LOCKS_bm_invisible_workers);
  /* move up to cp with worker_p */
  do {
    if (! move_up_one_node(NULL)) {
      return TRUE;
    }
  } while (! BITMAP_member(OrFr_members(LOCAL_top_or_fr), worker_p));
  /* put workers visibles */
  LOCK(GLOBAL_LOCKS_bm_invisible_workers);
  BITMAP_delete(GLOBAL_bm_invisible_workers, worker_id);
  BITMAP_delete(GLOBAL_bm_invisible_workers, worker_p);
  UNLOCK(GLOBAL_LOCKS_bm_invisible_workers);
  return (q_share_work(worker_p));
}


static int find_a_better_position(void){
  int i;
  bitmap busy_above, idle_above;
  BITMAP_difference(busy_above, GLOBAL_bm_present_workers, OrFr_members(LOCAL_top_or_fr));
  BITMAP_copy(idle_above, busy_above); 
  BITMAP_minus(busy_above, GLOBAL_bm_idle_workers);
  BITMAP_and(idle_above, GLOBAL_bm_idle_workers);
  for (i = 0; i < number_workers; i++) {
    if (BITMAP_member(idle_above, i)) {
      if (BITMAP_empty(busy_above)) 
        break;
      if (BITMAP_member(OrFr_members(REMOTE_top_or_fr(i)), worker_id))
        BITMAP_clear(busy_above);
      BITMAP_minus(busy_above, OrFr_members(REMOTE_top_or_fr(i)));
    }
  }
  if (BITMAP_empty(busy_above))
    return FALSE;
  /* move up to cp with all workers of bitmap busy_above */
  do {
    if (! move_up_one_node(NULL)) {
      return TRUE;
    }
  } while (! BITMAP_subset(OrFr_members(LOCAL_top_or_fr), busy_above));
  return FALSE;
}


static int search_for_hidden_shared_work(bitmap stable_busy){
  int i;
  bitmap invisible_work, idle_below;
  BITMAP_intersection(invisible_work, stable_busy, GLOBAL_bm_requestable_workers);
  BITMAP_intersection(idle_below, OrFr_members(LOCAL_top_or_fr), GLOBAL_bm_idle_workers);
  BITMAP_delete(idle_below, worker_id);
  for (i = 0; i < number_workers; i++) {
    if (BITMAP_member(idle_below ,i) && YOUNGER_CP(REMOTE_top_cp(i), LOCAL_top_cp))
      BITMAP_minus(invisible_work, OrFr_members(REMOTE_top_or_fr(i)));
  }
  if (BITMAP_empty(invisible_work))
    return FALSE;
  /* choose the first available worker */
  for (i = 0; i < number_workers; i++ ) {
    if (BITMAP_member(invisible_work ,i))
      break;
  }
  return (q_share_work(i));
}
#endif /* YAPOR */
