/* -------------------- **
**      Prototypes      **
** -------------------- */

STD_PROTO(static inline void unbind_variables, (tr_fr_ptr, tr_fr_ptr));
STD_PROTO(static inline void rebind_variables, (tr_fr_ptr, tr_fr_ptr));
STD_PROTO(static inline void restore_bindings, (tr_fr_ptr, tr_fr_ptr));



/* ------------------------- **
**      Inline funcions      **
** ------------------------- */

static inline
void unbind_variables(tr_fr_ptr start_tr, tr_fr_ptr end_tr) {
  CELL ref;

  /* unbind loop */
  while (start_tr != end_tr) {
    ref = (CELL) TrailTerm(--start_tr);
    /* check for global or local variables */
    if (IsVarTerm(ref)) {
      /* unbind variable */
      RESET_VARIABLE(ref);
    } else {
      ref = (CELL) RepPair(ref);
      if ((ADDR)ref >= TrailBase) {
        /* avoid frozen segments */
        start_tr = (tr_fr_ptr) ref;
#ifdef TABLING_ERRORS
        if (start_tr > (tr_fr_ptr) TrailTop)
          TABLING_ERROR_MESSAGE("start_tr > TrailTop (function unbind_variables)");
        if(start_tr < end_tr)
          TABLING_ERROR_MESSAGE("start_tr < end_tr (function unbind_variables)");
#endif /* TABLING_ERRORS */
      }
    }
  }
  return;
}


static inline
void rebind_variables(tr_fr_ptr start_tr, tr_fr_ptr end_tr) {
  CELL ref;

  /* rebind loop */
  while (start_tr != end_tr) {
    ref = (CELL) TrailTerm(--end_tr);
    /* check for global or local variables */
    if (IsVarTerm(ref)) {
      /* rebind variable */
      *((CELL *)ref) = TrailVal(end_tr);
    } else {
      ref = (CELL) RepPair(ref);
      if ((ADDR)ref >= TrailBase) {
        /* avoid frozen segments */
  	end_tr = (tr_fr_ptr) ref;
#ifdef TABLING_ERRORS
        if (end_tr > (tr_fr_ptr) TrailTop)
          TABLING_ERROR_MESSAGE("end_tr > TrailTop (function rebind_variables)");
        if(end_tr < start_tr)
          TABLING_ERROR_MESSAGE("end_tr < start_tr (function rebind_variables)");
#endif /* TABLING_ERRORS */
      }
    }
  }
  return;
}


static inline
void restore_bindings(tr_fr_ptr start_tr, tr_fr_ptr end_tr) {
  CELL ref;

  while (start_tr != end_tr) {
    /* unbind loop */
    while (start_tr > end_tr) {
      ref = (CELL) TrailTerm(--start_tr);
      if (IsVarTerm(ref)) {
        RESET_VARIABLE(ref);
      } else {
        ref = (CELL) RepPair(ref);
        if ((ADDR)ref >= TrailBase) {
          start_tr = (tr_fr_ptr) ref;
#ifdef TABLING_ERRORS
        if (start_tr > (tr_fr_ptr) TrailTop)
          TABLING_ERROR_MESSAGE("start_tr > TrailTop (function restore_bindings)");
#endif /* TABLING_ERRORS */
        }
      }
    }
    /* rebind loop */
    while (start_tr < end_tr) {
      ref = (CELL) TrailTerm(--end_tr);
      if (IsVarTerm(ref)) {
        *((CELL *)ref) = TrailVal(end_tr);
      } else {
        ref = (CELL) RepPair(ref);
        if ((ADDR)ref >= TrailBase) {
  	  end_tr = (tr_fr_ptr) ref;
#ifdef TABLING_ERRORS
        if (end_tr > (tr_fr_ptr) TrailTop)
          TABLING_ERROR_MESSAGE("end_tr > TrailTop (function restore_bindings)");
#endif /* TABLING_ERRORS */
        }
      }
    }
  }
  return;
}



/* ----------------------- **
**     Tabling Macros      **
** ----------------------- */

#ifdef YAPOR
#define find_leader_node(SG_FR, LEADER_CP, DEP_ON_STACK)                          \
        { dep_fr_ptr chain_dep_fr;                                                \
          /* find direct dependency node */                                       \
          if (SgFr_gen_worker(SG_FR) == worker_id) {                              \
            LEADER_CP = NORM_CP(SgFr_gen_cp(SG_FR));                              \
            DEP_ON_STACK = TRUE;                                                  \
          } else {                                                                \
            or_fr_ptr aux_or_fr = SgFr_gen_top_or_fr(SG_FR);                      \
            while (! BITMAP_member(OrFr_members(aux_or_fr), worker_id) &&         \
                   ! BITMAP_member(OrFr_hidden_members(aux_or_fr), worker_id))    \
              aux_or_fr = OrFr_next(aux_or_fr);                                   \
            LEADER_CP = OrFr_node(aux_or_fr);                                     \
            if (LEADER_CP == NORM_CP(SgFr_gen_cp(SG_FR))) DEP_ON_STACK = TRUE;    \
	    else DEP_ON_STACK = FALSE;                                            \
	  }                                                                       \
          /* find leader node */                                                  \
          chain_dep_fr = LOCAL_top_dep_fr;                                        \
          while (YOUNGER_CP(NORM_CP(DepFr_cons_cp(chain_dep_fr)), LEADER_CP)) {   \
            if (EQUAL_OR_YOUNGER_CP(LEADER_CP, DepFr_leader_cp(chain_dep_fr))) {  \
              LEADER_CP = DepFr_leader_cp(chain_dep_fr);                          \
              DEP_ON_STACK = DepFr_leader_dep_is_on_stack(chain_dep_fr);          \
              break;                                                              \
            }                                                                     \
            chain_dep_fr = DepFr_next(chain_dep_fr);                              \
          }                                                                       \
	}
#else
#define find_leader_node(SG_FR, LEADER_CP, DEP_ON_STACK)                          \
        { dep_fr_ptr chain_dep_fr;                                                \
          /* find direct dependency node */                                       \
          LEADER_CP = NORM_CP(SgFr_gen_cp(SG_FR));                                \
          DEP_ON_STACK = TRUE;                                                    \
          /* find leader node */                                                  \
          chain_dep_fr = LOCAL_top_dep_fr;                                        \
          while (YOUNGER_CP(NORM_CP(DepFr_cons_cp(chain_dep_fr)), LEADER_CP)) {   \
            if (EQUAL_OR_YOUNGER_CP(LEADER_CP, DepFr_leader_cp(chain_dep_fr))) {  \
              LEADER_CP = DepFr_leader_cp(chain_dep_fr);                          \
              break;                                                              \
            }                                                                     \
            chain_dep_fr = DepFr_next(chain_dep_fr);                              \
          }                                                                       \
	}
#endif /* YAPOR */


#ifdef YAPOR
#define YAPOR_SET_LOAD(CP_PTR)  SCH_set_load(CP_PTR)
#define SgFr_init_yapor_fields(SG_FR)                             \
        SgFr_gen_worker(SG_FR) = worker_id;                       \
        SgFr_gen_top_or_fr(SG_FR) = LOCAL_top_or_fr
#define DepFr_init_yapor_fields(DEP_FR, DEP_ON_STACK, TOP_OR_FR)  \
        DepFr_leader_dep_is_on_stack(DEP_FR) = DEP_ON_STACK;      \
        DepFr_top_or_fr(DEP_FR) = TOP_OR_FR
#define TrNode_init_yapor_fields(NODE, LTT, CUT, SEQ)             \
        TrNode_ltt(NODE) = LTT;                                   \
        TrNode_cut(NODE) = CUT;                                   \
        TrNode_seq(NODE) = SEQ
#else
#define YAPOR_SET_LOAD(CP_PTR)
#define SgFr_init_yapor_fields(SG_FR)
#define DepFr_init_yapor_fields(DEP_FR, DEP_ON_STACK, TOP_OR_FR)
#define TrNode_init_yapor_fields(NODE, LTT, CUT, SEQ)
#endif /* YAPOR */


#define new_suspension_frame(SUSP_FR, TOP_CP, TOP_OR_FR_ON_STACK, TOP_DEP, TOP_SG,  \
                             H_REG, B_REG, TR_REG, H_SIZE, B_SIZE, TR_SIZE)         \
        ALLOC_SUSPENSION_FRAME(SUSP_FR);                                            \
        SuspFr_top_cp(SUSP_FR) = TOP_CP;                                            \
        SuspFr_top_or_fr_on_stack(SUSP_FR) = TOP_OR_FR_ON_STACK;                    \
        SuspFr_top_dep_fr(SUSP_FR) = TOP_DEP;                                       \
        SuspFr_top_sg_fr(SUSP_FR) = TOP_SG;                                         \
        SuspFr_global_reg(SUSP_FR) = (void *) H_REG;                                \
        SuspFr_local_reg(SUSP_FR) = (void *) B_REG;                                 \
        SuspFr_trail_reg(SUSP_FR) = (void *) TR_REG;                                \
        SuspFr_global_start(SUSP_FR) = alloc_suspension_block(H_SIZE);              \
        SuspFr_local_start(SUSP_FR) = alloc_suspension_block(B_SIZE);               \
        SuspFr_trail_start(SUSP_FR) = alloc_suspension_block(TR_SIZE);              \
        SuspFr_global_size(SUSP_FR) = H_SIZE;                                       \
        SuspFr_local_size(SUSP_FR) = B_SIZE;                                        \
        SuspFr_trail_size(SUSP_FR) = TR_SIZE


#define new_subgoal_frame(SG_FR, BOTTOM_SG_NODE, GEN_CP, NEXT)              \
        { register ans_node_ptr ans_node;                                   \
          ALLOC_SUBGOAL_FRAME(SG_FR);                                       \
          INIT_LOCK(SgFr_lock(SG_FR));                                      \
          SgFr_gen_cp(SG_FR) = GEN_CP;                                      \
          SgFr_subgoal_trie(SG_FR) = BOTTOM_SG_NODE;                        \
          new_answer_trie_node(ans_node, 0, 0, 0, 0, 0, NULL, NULL, NULL);  \
          SgFr_answer_trie(SG_FR) = ans_node;                               \
          SgFr_first_answer(SG_FR) = NULL;                                  \
          SgFr_last_answer(SG_FR) = NULL;                                   \
          SgFr_completed(SG_FR) = FALSE;                                    \
          SgFr_init_yapor_fields(SG_FR);                                    \
          SgFr_next(SG_FR) = NEXT;                                          \
	}


#define new_dependency_frame(DEP_FR, DEP_ON_STACK, TOP_OR_FR, LEADER_CP, CONS_CP, SG_FR, NEXT)  \
        ALLOC_DEPENDENCY_FRAME(DEP_FR);                                                         \
        INIT_LOCK(DepFr_lock(DEP_FR));                                                          \
        DepFr_init_yapor_fields(DEP_FR, DEP_ON_STACK, TOP_OR_FR);                               \
        DepFr_leader_cp(DEP_FR) = LEADER_CP;                                                    \
        DepFr_cons_cp(DEP_FR) = CONS_CP;                                                        \
        DepFr_sg_fr(DEP_FR) = SG_FR;                                                            \
        DepFr_last_ans(DEP_FR) = NULL;                                                          \
        DepFr_h_fz(DEP_FR) = H_FZ;                                                              \
        DepFr_b_fz(DEP_FR) = B_FZ;                                                              \
        DepFr_tr_fz(DEP_FR) = TR_FZ;                                                            \
        DepFr_backchain_cp(DEP_FR) = NULL;                                                      \
        DepFr_next(DEP_FR) = NEXT


#define new_table_entry(TE, SUBGOAL_TRIE)       \
        ALLOC_TABLE_ENTRY(TE);                  \
        TabEnt_subgoal_trie(TE) = SUBGOAL_TRIE


#define new_subgoal_trie_node(NODE, ENTRY, CHILD, PARENT, PREVIOUS, NEXT)  \
        ALLOC_SUBGOAL_TRIE_NODE(NODE);                                     \
        INIT_LOCK(TrNode_lock(NODE));                                      \
        TrNode_entry(NODE) = ENTRY;                                        \
        TrNode_child(NODE) = CHILD;                                        \
        TrNode_parent(NODE) = PARENT;                                      \
        TrNode_previous(NODE) = PREVIOUS;                                  \
        TrNode_next(NODE) = NEXT


#define new_answer_trie_node(NODE, INSTR, LTT, CUT, SEQ, ENTRY, CHILD, PARENT, NEXT)  \
        ALLOC_ANSWER_TRIE_NODE(NODE);                                                 \
        INIT_LOCK(TrNode_lock(NODE));                                                 \
        TrNode_instr(NODE) = opcode(INSTR);                                           \
        TrNode_init_yapor_fields(NODE, LTT, CUT, SEQ);                                \
        TrNode_op_instr(NODE) = INSTR;                                                \
        TrNode_entry(NODE) = ENTRY;                                                   \
        TrNode_child(NODE) = CHILD;                                                   \
        TrNode_parent(NODE) = PARENT;                                                 \
        TrNode_next(NODE) = NEXT


#ifdef TAGS_FAST_OPS  /* Tags_32Ops.h */
#define	TabTagBits             TagBits
#define TabNumberOfLowTagBits  LowTagBits
#define	TabVarTagBits          MKTAG(0x0,0)
#define	TabAtomTagBits         AtomTag
#define	TabNumberTagBits       NumberTag
#define	TabPairTagBits         MKTAG(0x5,3)
#define	TabApplTagBits         MKTAG(0x5,0)
#else
#define	TabTagBits             OOOOPPS!!! Tabling inconsistent tag scheme
#define TabNumberOfLowTagBits  OOOOPPS!!! Tabling inconsistent tag scheme
#define	TabVarTagBits          OOOOPPS!!! Tabling inconsistent tag scheme
#define	TabAtomTagBits         OOOOPPS!!! Tabling inconsistent tag scheme
#define	TabNumberTagBits       OOOOPPS!!! Tabling inconsistent tag scheme
#define	TabPairTagBits         OOOOPPS!!! Tabling inconsistent tag scheme
#define	TabApplTagBits         OOOOPPS!!! Tabling inconsistent tag scheme
#endif


#define NORM_CP(CP)  ((choiceptr)(CP))
#define GEN_CP(CP)   ((gen_cp_ptr)(CP))
#define CONS_CP(CP)  ((cons_cp_ptr)(CP))


#define TAG_AS_ANSWER_LEAF_NODE(NODE)  ((unsigned int)TrNode_parent(NODE) |= 0x1)
#define UNTAG_ANSWER_LEAF_NODE(NODE)   ((ans_node_ptr)((unsigned int)NODE & 0xfffffffe))
#define IS_ANSWER_LEAF_NODE(NODE)      ((unsigned int)TrNode_parent(NODE) & 0x1)


#define FREE_STACK_PUSH(ITEM, STACK)        *--STACK = (CELL)(ITEM)
#define STACK_TOP(STACK)                    *STACK
#define STACK_POP(STACK)                    *STACK++
#define STACK_EMPTY(STACK, STACK_BASE)      STACK == STACK_BASE
#define STACK_NOT_EMPTY(STACK, STACK_BASE)  STACK != STACK_BASE
#define STACK_PUSH(ITEM, STACK, STACK_TOP)              \
        *--STACK = (CELL)(ITEM);                        \
        if (STACK <= STACK_TOP)                         \
          abort_optyap("auxiliary stack full")


#define MakeTableVarTerm(INDEX)    (INDEX << TabNumberOfLowTagBits)
#define VarIndexOfTableTerm(TERM)  (TERM >> TabNumberOfLowTagBits)
#define VarIndexOfTerm(TERM)       ((((CELL) TERM) - GLOBAL_table_var_enumerator(0))/sizeof(CELL))
#define IsTableVarTerm(TERM)       ((CELL) TERM) >= GLOBAL_table_var_enumerator(0) &&            \
                                   ((CELL) TERM) <= GLOBAL_table_var_enumerator(MAX_TABLE_VARS -1)
