/* ------------------ **
**      Includes      **
** ------------------ */

#include "Yap.h"
#ifdef TABLING
#include "tab.macros.h"
#ifdef YAPOR
#include "or.macros.h"
#endif /* YAPOR */



/* ------------------------------------- **
**      Local functions declaration      **
** ------------------------------------- */

#ifdef YAPOR
void release_suspension_branch(susp_fr_ptr susp_fr, dep_fr_ptr *chain_dep_fr, or_fr_ptr *chain_or_fr);
#endif /* YAPOR */



/* -------------------------- **
**      Global functions      **
** -------------------------- */

void private_completion(void) {
  sg_fr_ptr top_sg_fr;
  dep_fr_ptr dep_fr;

  TABLING_DEBUG_MESSAGE("Performing private completion");
  /* complete subgoals */
  top_sg_fr = SgFr_next(GEN_CP(B)->gcp_sg_fr);
  do {
    SgFr_completed(LOCAL_top_sg_fr) = TRUE;
    LOCAL_top_sg_fr = SgFr_next(LOCAL_top_sg_fr);
  } while (LOCAL_top_sg_fr != top_sg_fr);
  /* release dependency frames */
  while (YOUNGER_CP(NORM_CP(DepFr_cons_cp(LOCAL_top_dep_fr)), B)) {
    dep_fr = DepFr_next(LOCAL_top_dep_fr);
    FREE_DEPENDENCY_FRAME(LOCAL_top_dep_fr);
    LOCAL_top_dep_fr = dep_fr;
  }
  /* adjust registers */
  H_FZ = DepFr_h_fz(LOCAL_top_dep_fr);
  B_FZ = DepFr_b_fz(LOCAL_top_dep_fr);
  TR_FZ = DepFr_tr_fz(LOCAL_top_dep_fr);
  return;
}


#ifdef YAPOR
void public_completion(void) {
  dep_fr_ptr chain_dep_fr, next_dep_fr;
  or_fr_ptr chain_or_fr, top_or_fr, next_or_fr;
  susp_fr_ptr chain_susp_fr, next_susp_fr;
  sol_fr_ptr solutions, aux_solutions;

  /* Public completion only occurs in shared nodes. */
  /* This means that B and LOCAL_top_cp are equal. */
  OPTYAP_DEBUG_MESSAGE("Performing public completion");
  if (YOUNGER_CP(LOCAL_top_cp, B_FZ)) {
    /* The current node is a leader (and generator) node without younger */
    /* consumer nodes. So, we only have the current node to complete. */

    sg_fr_ptr top_sg_fr;
    /* complete subgoals */
    top_sg_fr = SgFr_next(GEN_CP(LOCL_top_cp)->gcp_sg_fr);



printf("***** DANGER public completion without younger consumer nodes\n");



    do {
      SgFr_completed(LOCAL_top_sg_fr) = TRUE;
      LOCAL_top_sg_fr = SgFr_next(LOCAL_top_sg_fr);
    } while (LOCAL_top_sg_fr != top_sg_fr);
    /* no dependency frames to release */
    chain_dep_fr = NULL;
    /* no need to adjust freeze registers */
  } else {
    /* The current node is a leader node with younger consumer nodes. */
    /* We need to complete all the dependent subgoals. */

    /* complete subgoals */
    if (DepFr_leader_dep_is_on_stack(LOCAL_top_dep_fr)) {
      while (LOCAL_top_sg_fr && EQUAL_OR_YOUNGER_CP(NORM_CP(SgFr_gen_cp(LOCAL_top_sg_fr)), LOCAL_top_cp)) {
        SgFr_completed(LOCAL_top_sg_fr) = TRUE;
        LOCAL_top_sg_fr = SgFr_next(LOCAL_top_sg_fr);
      }
    } else {
      while (LOCAL_top_sg_fr && YOUNGER_CP(NORM_CP(SgFr_gen_cp(LOCAL_top_sg_fr)), LOCAL_top_cp)) {
        SgFr_completed(LOCAL_top_sg_fr) = TRUE;
        LOCAL_top_sg_fr = SgFr_next(LOCAL_top_sg_fr);
      }
    }
    /* chain dependency frames to release */
    chain_dep_fr = NULL;
    while (YOUNGER_CP(NORM_CP(DepFr_cons_cp(LOCAL_top_dep_fr)), LOCAL_top_cp)) {
      LOCK(DepFr_lock(LOCAL_top_dep_fr));
      next_dep_fr = DepFr_next(LOCAL_top_dep_fr);
      DepFr_next(LOCAL_top_dep_fr) = chain_dep_fr;
      chain_dep_fr = LOCAL_top_dep_fr;
      LOCAL_top_dep_fr = next_dep_fr;
    }
    /* adjust freeze registers */
    H_FZ = DepFr_h_fz(LOCAL_top_dep_fr);
    B_FZ = DepFr_b_fz(LOCAL_top_dep_fr);
    TR_FZ = DepFr_tr_fz(LOCAL_top_dep_fr);
  }
  /* chain or-frames to release */
  chain_or_fr = NULL;
  top_or_fr = LOCAL_top_cp_on_stack->cp_or_fr;
  while (top_or_fr != LOCAL_top_or_fr) {
    LOCK_OR_FRAME(top_or_fr);
    chain_susp_fr = OrFr_suspensions(top_or_fr);
    while (chain_susp_fr) {
      release_suspension_branch(chain_susp_fr, &chain_dep_fr, &chain_or_fr);
      next_susp_fr = SuspFr_next(chain_susp_fr);
      FREE_SUSPENSION_FRAME(chain_susp_fr);
      chain_susp_fr = next_susp_fr;
    }
    OrFr_next(top_or_fr) = chain_or_fr;
    chain_or_fr = top_or_fr;
    top_or_fr = OrFr_next_on_stack(top_or_fr);
  }
  LOCK_OR_FRAME(top_or_fr);
  chain_susp_fr = OrFr_suspensions(top_or_fr);
  while (chain_susp_fr) {
    release_suspension_branch(chain_susp_fr, &chain_dep_fr, &chain_or_fr);
    next_susp_fr = SuspFr_next(chain_susp_fr);
    FREE_SUSPENSION_FRAME(chain_susp_fr);
    chain_susp_fr = next_susp_fr;
  }
  OrFr_suspensions(top_or_fr) = NULL;
  UNLOCK_OR_FRAME(top_or_fr);
  /* release dependency frames */
  while (chain_dep_fr) {
    next_dep_fr = DepFr_next(chain_dep_fr);
    FREE_DEPENDENCY_FRAME(chain_dep_fr);
    chain_dep_fr = next_dep_fr;
  }
  /* release or frames */
  solutions = NULL;
  while (chain_or_fr) {
    aux_solutions = OrFr_solutions(chain_or_fr);
    if (aux_solutions) {
      CUT_put_solutions_together(aux_solutions);
      SolFr_next(aux_solutions) = solutions;
      solutions = aux_solutions;
    }
    next_or_fr = OrFr_next(chain_or_fr);
    FREE_OR_FRAME(chain_or_fr);
    chain_or_fr = next_or_fr;
  }
  if (solutions) {
    CUT_put_solutions_together(solutions);
    SolFr_next(solutions) = OrFr_solutions(LOCAL_top_or_fr);
    OrFr_solutions(LOCAL_top_or_fr) = solutions;
  }
  /* adjust top register */
  LOCAL_top_cp_on_stack = LOCAL_top_cp;
  return;
}


void complete_suspension_frames(susp_fr_ptr chain_susp_fr) {
  dep_fr_ptr chain_dep_fr, next_dep_fr;
  or_fr_ptr chain_or_fr, next_or_fr;
  susp_fr_ptr next_susp_fr;
  sol_fr_ptr solutions, aux_solutions;



printf("   W%d: complete suspension frames na or-frame %x\n",worker_id,LOCAL_top_or_fr);



  OPTYAP_DEBUG_MESSAGE("Completing suspension frames in node %x", LOCAL_top_cp);
  solutions = NULL;
  chain_dep_fr = NULL;
  chain_or_fr = NULL;
  /* complete suspension frames */
  do {
    release_suspension_branch(chain_susp_fr, &chain_dep_fr, &chain_or_fr);
    next_susp_fr = SuspFr_next(chain_susp_fr);
    FREE_SUSPENSION_FRAME(chain_susp_fr);
    chain_susp_fr = next_susp_fr;
  } while (chain_susp_fr);
  /* release dependency frames */
  while (chain_dep_fr) {
    next_dep_fr = DepFr_next(chain_dep_fr);
    FREE_DEPENDENCY_FRAME(chain_dep_fr);
    chain_dep_fr = next_dep_fr;
  }
  /* release or frames */
  while (chain_or_fr) {
    aux_solutions = OrFr_solutions(chain_or_fr);
    if (aux_solutions) {
      CUT_put_solutions_together(aux_solutions);
      SolFr_next(aux_solutions) = solutions;
      solutions = aux_solutions;
    }
    next_or_fr = OrFr_next(chain_or_fr);
    FREE_OR_FRAME(chain_or_fr);
    chain_or_fr = next_or_fr;
  }
  if (solutions) {
    CUT_put_solutions_together(solutions);
    SolFr_next(solutions) = OrFr_solutions(LOCAL_top_or_fr);
    OrFr_solutions(LOCAL_top_or_fr) = solutions;
  }
  return;
}


void suspend_completion_branch(void) {
  susp_fr_ptr new_susp_fr;
  or_fr_ptr or_frame;
  long h_size, b_size, tr_size;

  /* Suspension operations only occurs in shared nodes. */
  /* This means that B and LOCAL_top_cp are equal. */
  OPTYAP_DEBUG_MESSAGE("Suspending current branch");
#ifdef OPTYAP_ERRORS
  if (B_FZ == LOCAL_top_cp)
    OPTYAP_ERROR_MESSAGE("B_FZ = LOCAL_top_cp (suspend_completion_branch)");
  if (YOUNGER_CP(LOCAL_top_cp, LOCAL_top_cp_on_stack))
    OPTYAP_ERROR_MESSAGE("YOUNGER_CP(LOCAL_top_cp, LOCAL_top_cp_on_stack) (suspend_completion_branch)");
  if (LOCAL_top_cp->cp_or_fr != LOCAL_top_or_fr)
    OPTYAP_ERROR_MESSAGE("LOCAL_top_cp->cp_or_fr != LOCAL_top_or_fr (suspend_completion_branch)");
  or_frame = LOCAL_top_cp_on_stack->cp_or_fr;
  while (or_frame != LOCAL_top_or_fr) {
    if (YOUNGER_CP(LOCAL_top_cp, OrFr_node(or_frame))) {
      OPTYAP_ERROR_MESSAGE("YOUNGER_CP(LOCAL_top_cp, OrFr_node(or_frame)) (suspend_completion_branch)");
      break;
    }
    or_frame = OrFr_next_on_stack(or_frame);
  }
#endif /* OPTYAP_ERRORS */
  if (YOUNGER_CP(LOCAL_top_cp, B_FZ)) {
    /* The current node is a leader (and generator) node without younger */
    /* consumer nodes. So, we only have the current node to suspend. */
    new_suspension_frame(new_susp_fr, LOCAL_top_cp, LOCAL_top_cp_on_stack->cp_or_fr, 
                         LOCAL_top_dep_fr, LOCAL_top_sg_fr,
                         NULL, NULL, NULL, 0, 0, 0);
    /* adjust top pointers */
    if (SgFr_gen_worker(LOCAL_top_sg_fr) == worker_id)
      SgFr_gen_worker(LOCAL_top_sg_fr) = number_workers;
    LOCAL_top_sg_fr = SgFr_next(LOCAL_top_sg_fr);
  } else {
    /* The current node is a leader node with younger consumer nodes. */
    /* We need to suspend all the dependent branches. */
    h_size = (unsigned long) H_FZ - (unsigned long) LOCAL_top_cp->cp_h;
    b_size = (unsigned long) LOCAL_top_cp - (unsigned long) B_FZ;
    tr_size = (unsigned long) TR_FZ - (unsigned long) LOCAL_top_cp->cp_tr;
    new_suspension_frame(new_susp_fr, LOCAL_top_cp, LOCAL_top_cp_on_stack->cp_or_fr,
                         LOCAL_top_dep_fr, LOCAL_top_sg_fr,
                         LOCAL_top_cp->cp_h, B_FZ, LOCAL_top_cp->cp_tr, h_size, b_size, tr_size);
    memcpy(SuspFr_global_start(new_susp_fr),
           SuspFr_global_reg(new_susp_fr),
           h_size);
    memcpy(SuspFr_local_start(new_susp_fr),
           SuspFr_local_reg(new_susp_fr),
           b_size);
    memcpy(SuspFr_trail_start(new_susp_fr),
           SuspFr_trail_reg(new_susp_fr),
           tr_size);
    /* update shared nodes */
    or_frame = LOCAL_top_cp_on_stack->cp_or_fr;
    while (or_frame != LOCAL_top_or_fr) {
      LOCK_OR_FRAME(or_frame);
      BITMAP_insert(OrFr_hidden_members(or_frame), number_workers);
      BITMAP_delete(OrFr_hidden_members(or_frame), worker_id);
      UNLOCK_OR_FRAME(or_frame);
      or_frame = OrFr_next_on_stack(or_frame);
    }
    /* adjust top pointers */
    if (DepFr_leader_dep_is_on_stack(LOCAL_top_dep_fr)) {
      while (LOCAL_top_sg_fr && EQUAL_OR_YOUNGER_CP(NORM_CP(SgFr_gen_cp(LOCAL_top_sg_fr)), LOCAL_top_cp)) {
        if (SgFr_gen_worker(LOCAL_top_sg_fr) == worker_id)
          SgFr_gen_worker(LOCAL_top_sg_fr) = number_workers;
        LOCAL_top_sg_fr = SgFr_next(LOCAL_top_sg_fr);
      }
    } else {
      while (LOCAL_top_sg_fr && YOUNGER_CP(NORM_CP(SgFr_gen_cp(LOCAL_top_sg_fr)), LOCAL_top_cp)) {
        if (SgFr_gen_worker(LOCAL_top_sg_fr) == worker_id)
          SgFr_gen_worker(LOCAL_top_sg_fr) = number_workers;
        LOCAL_top_sg_fr = SgFr_next(LOCAL_top_sg_fr);
      }
    }
    while (YOUNGER_CP(NORM_CP(DepFr_cons_cp(LOCAL_top_dep_fr)), LOCAL_top_cp))
      LOCAL_top_dep_fr = DepFr_next(LOCAL_top_dep_fr);
    LOCAL_top_cp_on_stack = LOCAL_top_cp;
    /* adjust frozen registers */
    H_FZ = DepFr_h_fz(LOCAL_top_dep_fr);
    B_FZ = DepFr_b_fz(LOCAL_top_dep_fr);
    TR_FZ = DepFr_tr_fz(LOCAL_top_dep_fr);
  }
  /* store suspension frame in current top or-frame */
  LOCK_OR_FRAME(LOCAL_top_or_fr);
  SuspFr_next(new_susp_fr) = OrFr_suspensions(LOCAL_top_or_fr);
  OrFr_suspensions(LOCAL_top_or_fr) = new_susp_fr;
  UNLOCK_OR_FRAME(LOCAL_top_or_fr);
#ifdef OPTYAP_ERRORS
  or_frame = SuspFr_top_or_fr_on_stack(new_susp_fr);
  while (or_frame != SuspFr_top_cp(new_susp_fr)->cp_or_fr) {
    if (YOUNGER_CP(SuspFr_top_cp(new_susp_fr), OrFr_node(or_frame))) {
      OPTYAP_ERROR_MESSAGE("YOUNGER_CP(SuspFr_top_cp(new_susp_fr), OrFr_node(or_frame)) (suspend_completion_branch)");
      break;
    }
    or_frame = OrFr_next_on_stack(or_frame);
  }
  if (LOCAL_top_or_fr != SuspFr_top_cp(new_susp_fr)->cp_or_fr)
    OPTYAP_ERROR_MESSAGE("LOCAL_top_or_fr != SuspFr_top_cp(new_susp_fr)->cp_or_fr (suspend_completion_branch)");
#endif /* OPTYAP_ERRORS */



printf("   W%d: suspendind susp_fr %x na or_frame %x\n",worker_id,new_susp_fr, LOCAL_top_or_fr);



  return;
}


void resume_suspension_frame(susp_fr_ptr resume_fr) {
  or_fr_ptr or_frame;
  sg_fr_ptr sg_frame;


printf("resuming frame %x\n",resume_fr);



  OPTYAP_DEBUG_MESSAGE("Resuming suspension frame %x", resume_fr);
#ifdef OPTYAP_ERRORS
  if (LOCAL_top_cp != SuspFr_top_cp(resume_fr))
    OPTYAP_ERROR_MESSAGE("LOCAL_top_cp != SuspFr_top_cp (resume_suspension_frame)");
  if (LOCAL_top_cp != SuspFr_local_reg(resume_fr) + SuspFr_local_size(resume_fr))
    OPTYAP_ERROR_MESSAGE("LOCAL_top_cp != SuspFr_local_reg + SuspFr_local_size (resume_suspension_frame)");

  if (LOCAL_top_cp->cp_h != SuspFr_global_reg(resume_fr))
    OPTYAP_ERROR_MESSAGE("LOCAL_top_cp->cp_h != SuspFr_global_reg (resume_suspension_frame)");
  if (DepFr_h_fz(SuspFr_top_dep_fr(resume_fr)) != SuspFr_global_reg(resume_fr) + SuspFr_global_size(resume_fr))
    OPTYAP_ERROR_MESSAGE("DepFr_h_fz(SuspFr_top_dep_fr) != SuspFr_global_reg + Suspfr_global_size (resume_suspension_frame)");

  if (LOCAL_top_cp->cp_tr != SuspFr_trail_reg(resume_fr))
    OPTYAP_ERROR_MESSAGE("LOCAL_top_cp != SuspFr_local_reg (resume_suspension_frame)");
  if (DepFr_tr_fz(SuspFr_top_dep_fr(resume_fr)) != SuspFr_trail_reg(resume_fr) + SuspFr_trail_size(resume_fr))
    OPTYAP_ERROR_MESSAGE("DepFr_tr_fz(SuspFr_top_dep_fr) != SuspFr_trail_reg + SuspFr_trail_size (resume_suspension_frame)");
#endif /* OPTYAP_ERRORS */
  /* resume suspension branch */
  memcpy(SuspFr_global_reg(resume_fr),
         SuspFr_global_start(resume_fr),
         SuspFr_global_size(resume_fr));
  memcpy(SuspFr_local_reg(resume_fr),
         SuspFr_local_start(resume_fr),
         SuspFr_local_size(resume_fr));
  memcpy(SuspFr_trail_reg(resume_fr),
         SuspFr_trail_start(resume_fr),
         SuspFr_trail_size(resume_fr));
#ifdef OPTYAP_ERRORS
  if (DepFr_cons_cp(SuspFr_top_dep_fr(resume_fr))->ccp_h != SuspFr_global_reg(resume_fr) + SuspFr_global_size(resume_fr))
    OPTYAP_ERROR_MESSAGE("DepFr_cons_cp(SuspFr_top_dep_fr)->cp_h != SuspFr_global_reg + Suspfr_global_size (resume_suspension_frame)");

  if (DepFr_cons_cp(SuspFr_top_dep_fr(resume_fr))->ccp_tr != SuspFr_trail_reg(resume_fr) + SuspFr_trail_size(resume_fr))
    OPTYAP_ERROR_MESSAGE("DepFr_cons_cp(SuspFr_top_dep_fr)->cp_tr != SuspFr_trail_reg + SuspFr_trail_size (resume_suspension_frame)");
#endif /* OPTYAP_ERRORS */
  /* update shared nodes */
  or_frame = SuspFr_top_or_fr_on_stack(resume_fr);
  LOCAL_top_cp_on_stack = OrFr_node(or_frame);
  while (or_frame != SuspFr_top_cp(resume_fr)->cp_or_fr) {
    LOCK_OR_FRAME(or_frame);
    BITMAP_insert(OrFr_hidden_members(or_frame), worker_id);
    BITMAP_delete(OrFr_hidden_members(or_frame), number_workers);
    UNLOCK_OR_FRAME(or_frame);
    or_frame = OrFr_next_on_stack(or_frame);
  }
  while (or_frame != LOCAL_top_or_fr) {
    LOCK_OR_FRAME(or_frame);
    BITMAP_insert(OrFr_members(or_frame), worker_id);
    BITMAP_delete(OrFr_hidden_members(or_frame), worker_id);
    UNLOCK_OR_FRAME(or_frame);
    or_frame = OrFr_next(or_frame);
  }
  /* adjust top pointers */
  sg_frame = SuspFr_top_sg_fr(resume_fr);
  while (sg_frame != LOCAL_top_sg_fr) {
    SgFr_gen_worker(sg_frame) = worker_id;
    sg_frame = SgFr_next(sg_frame);
  }
  LOCAL_top_sg_fr = SuspFr_top_sg_fr(resume_fr);
  LOCAL_top_dep_fr = SuspFr_top_dep_fr(resume_fr);
  LOCAL_top_cp = SuspFr_top_cp(resume_fr);
  LOCAL_top_or_fr = LOCAL_top_cp->cp_or_fr;
  /* adjust frozen registers */
  H_FZ = DepFr_h_fz(LOCAL_top_dep_fr);
  B_FZ = DepFr_b_fz(LOCAL_top_dep_fr);
  TR_FZ = DepFr_tr_fz(LOCAL_top_dep_fr);
  /* free suspension frame */
  FREE_SUSPENSION_FRAME(resume_fr);
  return;
}


void collect_suspension_frames_with_unconsumed_answers(void) {
  susp_fr_ptr *susp_ptr, next_susp_fr, collected_susp_fr;
#ifdef OPTYAP_ERRORS
  if (IS_UNLOCKED(LOCAL_top_or_fr))
    OPTYAP_ERROR_MESSAGE("LOCAL_top_or_fr unlocked (collect_suspension_frames_with_unconsumed_answers)");
#endif /* OPTYAP_ERRORS */
  collected_susp_fr = NULL;
  susp_ptr = & OrFr_suspensions(LOCAL_top_or_fr);
  while ((next_susp_fr = *susp_ptr) != NULL) {
    /* check for unconsumed answers in suspension frame */
    if (suspension_frame_with_unconsumed_answers(next_susp_fr)) {
      /* collect suspension frame */
      OPTYAP_DEBUG_MESSAGE("Collecting suspension frame %x", next_susp_fr);
      *susp_ptr = SuspFr_next(next_susp_fr);
      SuspFr_next(next_susp_fr) = collected_susp_fr;
      collected_susp_fr = next_susp_fr;
    }
    susp_ptr = & SuspFr_next(*susp_ptr);
  }
  /* check for collected suspension frames */
  if (collected_susp_fr) {
    susp_ptr = & LOCAL_top_collected_susp_fr;
    /* find chain point to put the collected suspension frames */
    while (*susp_ptr && YOUNGER_CP(SuspFr_top_cp(*susp_ptr), LOCAL_top_cp)) {
      susp_ptr = & SuspFr_next(*susp_ptr);
    }
    /* chain the new collected suspension frames with the old ones */
    next_susp_fr = *susp_ptr;
    *susp_ptr = collected_susp_fr;
    while (SuspFr_next(collected_susp_fr)) {
      collected_susp_fr = SuspFr_next(collected_susp_fr);
    }
    SuspFr_next(collected_susp_fr) = next_susp_fr;
  }
  return;
}


int suspension_frame_with_unconsumed_answers(susp_fr_ptr susp_fr) {
  dep_fr_ptr dep_fr;
  choiceptr top_cp;

  dep_fr = SuspFr_top_dep_fr(susp_fr);
  top_cp = SuspFr_top_cp(susp_fr);
  while (YOUNGER_CP(NORM_CP(DepFr_cons_cp(dep_fr)), top_cp)) {
    if (DepFr_last_ans(dep_fr) != SgFr_last_answer(DepFr_sg_fr(dep_fr)))
      return TRUE;
    dep_fr = DepFr_next(dep_fr);
  }
  return FALSE;
}



/* ------------------------- **
**      Local functions      **
** ------------------------- */

void release_suspension_branch(susp_fr_ptr susp_fr, dep_fr_ptr *chain_dep_fr, or_fr_ptr *chain_or_fr) {
  sg_fr_ptr aux_sg_fr;
  dep_fr_ptr aux_dep_fr, next_dep_fr;
  or_fr_ptr top_or_fr;
  susp_fr_ptr chain_susp_fr, next_susp_fr;

  /* complete subgoals */
  aux_sg_fr = SuspFr_top_sg_fr(susp_fr);
  if (DepFr_leader_dep_is_on_stack(SuspFr_top_dep_fr(susp_fr))) {
    while (aux_sg_fr &&
           EQUAL_OR_YOUNGER_CP(NORM_CP(SgFr_gen_cp(aux_sg_fr)), SuspFr_top_cp(susp_fr)) &&
           ! SgFr_completed(aux_sg_fr)) {
      SgFr_completed(aux_sg_fr) = TRUE;
      aux_sg_fr = SgFr_next(aux_sg_fr);
    }
  } else {
    while (aux_sg_fr &&
           YOUNGER_CP(NORM_CP(SgFr_gen_cp(aux_sg_fr)), SuspFr_top_cp(susp_fr)) &&
           ! SgFr_completed(aux_sg_fr)) {
      SgFr_completed(aux_sg_fr) = TRUE;
      aux_sg_fr = SgFr_next(aux_sg_fr);
    }
  }
  /* chain dependency frames to release */
  aux_dep_fr = SuspFr_top_dep_fr(susp_fr);
  while (YOUNGER_CP(NORM_CP(DepFr_cons_cp(aux_dep_fr)), SuspFr_top_cp(susp_fr))
      && IS_UNLOCKED(DepFr_lock(aux_dep_fr))) {
    LOCK(DepFr_lock(aux_dep_fr));
    next_dep_fr = DepFr_next(aux_dep_fr);
    DepFr_next(aux_dep_fr) = *chain_dep_fr;
    *chain_dep_fr = aux_dep_fr;
    aux_dep_fr = next_dep_fr;
  }
  /* chain or-frames to release */
  top_or_fr = SuspFr_top_or_fr_on_stack(susp_fr);
  while (IS_UNLOCKED(OrFr_lock(top_or_fr))) {
#ifdef OPTYAP_ERRORS
    if (YOUNGER_CP(SuspFr_top_cp(susp_fr), OrFr_node(top_or_fr)))
      OPTYAP_ERROR_MESSAGE("YOUNGER_CP(SuspFr_top_cp(susp_fr), OrFr_node(top_or_fr)) (release_suspension_branch)");
#endif /* OPTYAP_ERRORS */
    LOCK_OR_FRAME(top_or_fr);
    chain_susp_fr = OrFr_suspensions(top_or_fr);
    while (chain_susp_fr) {
      release_suspension_branch(chain_susp_fr, chain_dep_fr, chain_or_fr);
      next_susp_fr = SuspFr_next(chain_susp_fr);
      FREE_SUSPENSION_FRAME(chain_susp_fr);
      chain_susp_fr = next_susp_fr;
    }
    OrFr_next(top_or_fr) = *chain_or_fr;
    *chain_or_fr = top_or_fr;
    top_or_fr = OrFr_next_on_stack(top_or_fr);
  }
  return;
}
#endif /* YAPOR */
#endif /* TABLING */
