/* ------------------ **
**      Includes      **
** ------------------ */

#include "Yap.h"
#ifdef TABLING
#include "Yatom.h"
#include "tab.macros.h"
#include <stdio.h>



/* ----------------------- **
**      Local inlines      **
** ----------------------- */

STD_PROTO(static inline sg_node_ptr subgoal_trie_node_check_insert, (sg_node_ptr, Term));
STD_PROTO(static inline ans_node_ptr answer_trie_node_check_insert, (ans_node_ptr, Term, int));

static inline
sg_node_ptr subgoal_trie_node_check_insert(sg_node_ptr parent_node, Term t) {
  sg_node_ptr child_node;
  LOCK(TrNode_lock(parent_node));
  child_node = TrNode_child(parent_node);
  if (child_node == NULL) {
    new_subgoal_trie_node(child_node, t, NULL, parent_node, NULL, NULL);
    TrNode_child(parent_node) = child_node;
  } else {
    while (child_node && TrNode_entry(child_node) != t) {
      child_node = TrNode_next(child_node);
    }
    if (child_node == NULL) {
      new_subgoal_trie_node(child_node, t, NULL, parent_node, NULL, TrNode_child(parent_node));
      TrNode_previous(TrNode_child(parent_node)) = child_node;
      TrNode_child(parent_node) = child_node;
    }
  }
  UNLOCK(TrNode_lock(parent_node));
  return child_node;    
}


static inline 
ans_node_ptr answer_trie_node_check_insert(ans_node_ptr parent_node, Term t, int instr) {
  ans_node_ptr child_node;
  LOCK(TrNode_lock(parent_node));
  child_node = TrNode_child(parent_node);
  if (child_node == NULL) {
    new_answer_trie_node(child_node, instr, 1, 0, 0, t, NULL, parent_node, NULL);
    TrNode_child(parent_node) = child_node;
  } else {
    while (child_node && TrNode_entry(child_node) != t) {
      child_node = TrNode_next(child_node);
    }
    if (child_node == NULL) {
      ans_node_ptr aux_child = TrNode_child(parent_node);
      new_answer_trie_node(child_node, instr + 2, TrNode_ltt(aux_child) + 1,
                           0, 0, t, NULL, parent_node, aux_child);
      TrNode_instr(aux_child) = opcode(++TrNode_op_instr(aux_child));
      TrNode_child(parent_node) = child_node;
    }
  }
  UNLOCK(TrNode_lock(parent_node));
  return child_node;
}



/* ------------------------------------- **
**      Local functions declaration      **
** ------------------------------------- */

static sg_fr_ptr chain_trie_subgoals(sg_node_ptr sg_node, int missing_nodes, sg_fr_ptr last_sg_fr);
static void free_subgoal_trie_branch(sg_node_ptr child_node);
static void free_answer_trie_branch(ans_node_ptr node);



/* -------------------------- **
**      Global functions      **
** -------------------------- */

sg_node_ptr subgoal_search(OPREG arity, sg_node_ptr current_sg_node, CELL **Yaddr) {
  int i, j, count_vars;
  CELL *stack_vars, *stack_terms_base, *stack_terms_top, *stack_terms;
  
  count_vars = 0;
  stack_vars = *Yaddr;
  stack_terms_top = (CELL *)TrailTop;
  stack_terms_base = stack_terms = AuxSp;

  for (i = 1; i <= arity; i++) {
    STACK_PUSH(XREGS[i], stack_terms, stack_terms_top);
    do {
      Term t = Deref(STACK_POP(stack_terms));
      int tag = t & TabTagBits;
      switch (tag) {
        case TabVarTagBits:
          if (IsTableVarTerm(t)) {
	    t = MakeTableVarTerm(VarIndexOfTerm(t));
            current_sg_node = subgoal_trie_node_check_insert(current_sg_node, t);
  	  } else {
            FREE_STACK_PUSH(t, stack_vars);
	    *((CELL *)t) = GLOBAL_table_var_enumerator(count_vars);
            t = MakeTableVarTerm(count_vars);
            count_vars++;
            current_sg_node = subgoal_trie_node_check_insert(current_sg_node, t);
	  }
          break;
        case TabAtomTagBits:
        case TabNumberTagBits:
          current_sg_node = subgoal_trie_node_check_insert(current_sg_node, t);
          break;
        case TabPairTagBits:
          current_sg_node = subgoal_trie_node_check_insert(current_sg_node, TabPairTagBits);
          STACK_PUSH(RepPair(t) + 1, stack_terms, stack_terms_top);
          STACK_PUSH(RepPair(t), stack_terms, stack_terms_top);
          break;
        case TabApplTagBits:
          current_sg_node = subgoal_trie_node_check_insert(current_sg_node, TAGGEDA(TabApplTagBits, FunctorOfTerm(t)));
          for (j = ArityOfFunctor(FunctorOfTerm(t)); j >= 1; j--)
            STACK_PUSH(RepAppl(t) + j, stack_terms, stack_terms_top);
          break;
        default :
          abort_optyap("unknown type tag in function subgoal_search");
      }
    } while (STACK_NOT_EMPTY(stack_terms, stack_terms_base));
  }

  FREE_STACK_PUSH(count_vars, stack_vars);
  *Yaddr = stack_vars++;
  /* restore variables */
  while (count_vars--)
    *((CELL *)*stack_vars) = STACK_POP(stack_vars);

  return current_sg_node;
}


ans_node_ptr answer_search(ans_node_ptr current_ans_node, CELL *subs_ptr) {
  int i, j, count_vars, subs_arity;
  CELL *stack_vars, *stack_terms_base, *stack_terms_top, *stack_terms;

  count_vars = 0;
  subs_arity = *subs_ptr;
  stack_vars = AuxSp;
  stack_terms_top = (CELL *)TrailTop;
  stack_terms_base = stack_terms = stack_vars - MAX_TABLE_VARS;

  for (i = subs_arity; i >= 1; i--) {
    STACK_PUSH(*(subs_ptr + i), stack_terms, stack_terms_top);
    do {
      Term t = Deref(STACK_POP(stack_terms));
      int tag = t & TabTagBits;
      switch (tag) {
        case TabVarTagBits:
          if (IsTableVarTerm(t)) {
	    t = MakeTableVarTerm(VarIndexOfTerm(t));
            current_ans_node = answer_trie_node_check_insert(current_ans_node, t, _trie_do_val);
  	  } else {
            FREE_STACK_PUSH(t, stack_vars);
	    *((CELL *)t) = GLOBAL_table_var_enumerator(count_vars);
            t = MakeTableVarTerm(count_vars);
            count_vars++;
            current_ans_node = answer_trie_node_check_insert(current_ans_node, t, _trie_do_var);
	  }
          break;
        case TabAtomTagBits:
        case TabNumberTagBits:
          current_ans_node = answer_trie_node_check_insert(current_ans_node, t, _trie_do_atom);
          break;
        case TabPairTagBits:
          current_ans_node = answer_trie_node_check_insert(current_ans_node, TabPairTagBits, _trie_do_list);
          STACK_PUSH(RepPair(t) + 1, stack_terms, stack_terms_top);
          STACK_PUSH(RepPair(t), stack_terms, stack_terms_top);
          break;
        case TabApplTagBits:
          current_ans_node = answer_trie_node_check_insert(current_ans_node,
                                    TAGGEDA(TabApplTagBits, FunctorOfTerm(t)),
                                    _trie_do_struct);
          for (j = ArityOfFunctor(FunctorOfTerm(t)); j >= 1; j--)
            STACK_PUSH(RepAppl(t) + j, stack_terms, stack_terms_top);
          break;
        default :
          abort_optyap("unknown type tag in function answer_search");
      }
    } while (STACK_NOT_EMPTY(stack_terms, stack_terms_base));
  }

  /* restore variables */
  while (count_vars--)
    *((CELL *)*stack_vars) = STACK_POP(stack_vars);

  return current_ans_node;
}


int p_check_incomplete_tables(void) {
  if (LOCAL_top_sg_fr != NULL)
    INFORMATION_MESSAGE("Unexpected incomplete subgoals");
  if (LOCAL_top_dep_fr != GLOBAL_root_dep_fr)
    INFORMATION_MESSAGE("Unexpected incomplete dependency frames");
  return(TRUE);
}


int p_abolish_incomplete_tables(void) {
  /* abolish incomplete subgoals */
  while (LOCAL_top_sg_fr != NULL) {
    sg_fr_ptr sg_fr;
    free_answer_trie_branch(SgFr_answer_trie(LOCAL_top_sg_fr));
    free_subgoal_trie_branch(SgFr_subgoal_trie(LOCAL_top_sg_fr));
    sg_fr = LOCAL_top_sg_fr;
    LOCAL_top_sg_fr = SgFr_next(LOCAL_top_sg_fr);
    TABLING_DEBUG_MESSAGE("Abolishing incomplete subgoal (sg_fr %x)", (unsigned int)sg_fr);
    FREE_SUBGOAL_FRAME(sg_fr);
  }
  /* release dependency frames */
  while (LOCAL_top_dep_fr != GLOBAL_root_dep_fr) {
    dep_fr_ptr dep_fr = DepFr_next(LOCAL_top_dep_fr);
    FREE_DEPENDENCY_FRAME(LOCAL_top_dep_fr);
    LOCAL_top_dep_fr = dep_fr;
  }
  /* adjust registers */
  H_FZ = DepFr_h_fz(LOCAL_top_dep_fr);
  B_FZ = DepFr_b_fz(LOCAL_top_dep_fr);
  TR_FZ = DepFr_tr_fz(LOCAL_top_dep_fr);
  return(TRUE);
}


int p_abolish_trie(void) {
  Term t;
  Atom at;
  int arity;
  PredEntry *pe;
  sg_fr_ptr chain_sg_fr;

  t = Deref(ARG1);
  if (IsAtomTerm(t)) {
    at = AtomOfTerm(t);
    arity = 0;
  } else if (IsApplTerm(t)) {
    Functor func = FunctorOfTerm(t);
    at = NameOfFunctor(func);
    arity = ArityOfFunctor(func);
  } else {
    abort_optyap("unknown term in function p_abolish_trie");
    at = NULL; /* just to avoid gcc warning */
    arity = 0; /* just to avoid gcc warning */
  }
  pe = RepPredProp(PredProp(at, arity));
  /* check for empty tries */
  if (TrNode_child(TabEnt_subgoal_trie(pe->TableOfPred))) {
    /* chain trie subgoals */
    chain_sg_fr = chain_trie_subgoals(TabEnt_subgoal_trie(pe->TableOfPred), arity, NULL);
    /* abolish subgoals */
    while (chain_sg_fr != NULL) {
      sg_fr_ptr aux_sg_fr;
      free_answer_trie_branch(SgFr_answer_trie(chain_sg_fr));
      free_subgoal_trie_branch(SgFr_subgoal_trie(chain_sg_fr));
      aux_sg_fr = chain_sg_fr;
      chain_sg_fr = SgFr_next(chain_sg_fr);
      TABLING_DEBUG_MESSAGE("Abolishing subgoal (sg_fr %x)", (unsigned int)aux_sg_fr);
      FREE_SUBGOAL_FRAME(aux_sg_fr);
    }
  }
  return (TRUE);
}


void show_trie(sg_node_ptr sg_node, int pred_arity, Atom pred_atom) {
  char str[1000];
  int arity[100];
  int str_index;

  str_index = sprintf(str, "  ?- %s(", AtomName(pred_atom));
  arity[0] = 1;
  arity[1] = pred_arity;
  fprintf(stdout, "\n[ Trie structure for predicate '%s/%d' ]\n[\n", AtomName(pred_atom), pred_arity);
  show_subgoal_trie(sg_node, str, str_index, arity);
  /*show_subgoal_trie(sg_node, pred_arity, 0);*/
  fprintf(stdout, "]\n\n");
}


void show_subgoal_trie(sg_node_ptr sg_node, char *str, int str_index, int *arity) {
  int tag;
  Term t;
  int new_arity[100];

  if (sg_node == NULL) return;

  if (arity[0] == 0) {
    ans_node_ptr ans_node;
    str[str_index] = 0;
    fprintf(stdout, "%s.\n", str);
    if (! SgFr_completed((sg_fr_ptr)sg_node)) {
      fprintf(stdout, "    POTENCIAL ERROR: SUBGOAL NOT COMPLETED !!!\n");
    }
    ans_node = SgFr_answer_trie((sg_fr_ptr)sg_node);
    if (IS_ANSWER_LEAF_NODE(ans_node)) {
      fprintf(stdout, "    YES\n");
    } else if (TrNode_child(ans_node) == NULL) {
      fprintf(stdout, "    NO\n");
    } else {
      char answer_str[1000];
      int answer_arity[1000];
      answer_arity[0] = 0;
      show_answer_trie(TrNode_child(ans_node), answer_str, 0, answer_arity, 0);
    }
    return;
  }
  memcpy(new_arity, arity, 100);
  show_subgoal_trie(TrNode_next(sg_node), str, str_index, new_arity);

  t = TrNode_entry(sg_node);
  tag = t & TabTagBits;
  switch (tag) {
    case TabVarTagBits:
      if (arity[arity[0]] == -1) str[str_index - 1] = '|';
      str_index += sprintf(& str[str_index], "VAR%d", VarIndexOfTableTerm(t));
      while (arity[0] && (arity[arity[0]] == 1 || arity[arity[0]] == -1)) {
        if(arity[arity[0]] == 1) {
          str_index += sprintf(& str[str_index], ")");
          arity[0]--;
	}
        if(arity[arity[0]] == -1) {
          str_index += sprintf(& str[str_index], "]");
          arity[0]--;
	}
      }
      if (arity[0]) {
        if (arity[arity[0]] == -2) arity[arity[0]] = -1;
        else arity[arity[0]]--;
        str_index += sprintf(& str[str_index], ",");
      }
      break;
    case TabNumberTagBits:
      str_index += sprintf(& str[str_index], "%d", IntOfTerm(t));
      while (arity[0] && arity[arity[0]] == 1) {
        str_index += sprintf(& str[str_index], ")");
        arity[0]--;
      }
      if (arity[0]) {
        if (arity[arity[0]] == -2) arity[arity[0]] = -1;
        else arity[arity[0]]--;
        str_index += sprintf(& str[str_index], ",");
      }
      break;
    case TabAtomTagBits:
      if (arity[arity[0]] == -1) {
        if (strcmp("[]",  AtomName(AtomOfTerm(t))))
          fprintf(stdout, "    POTENCIAL ERROR: PAIR WITHOUT END ATOM '[]' !!!\n");
        str[str_index - 1] = ']';
        arity[0]--;
      } else {
        str_index += sprintf(& str[str_index], "%s", AtomName(AtomOfTerm(t)));
      }
      while (arity[0] && arity[arity[0]] == 1) {
        str_index += sprintf(& str[str_index], ")");
        arity[0]--;
      }
      if (arity[0]) {
        if (arity[arity[0]] == -2) arity[arity[0]] = -1;
        else arity[arity[0]]--;
        str_index += sprintf(& str[str_index], ",");
      }
      break;
    case TabPairTagBits:
      if (arity[arity[0]] == -1) {
        arity[arity[0]] = -2;
      } else {
        str_index += sprintf(& str[str_index], "[");
        arity[0]++;
        arity[arity[0]] = -2;
      }
      break;
    case TabApplTagBits:
      str_index += sprintf(& str[str_index], "%s(", AtomName(NameOfFunctor((Functor)NonTagPart(t))));
      arity[0]++;
      arity[arity[0]] = ArityOfFunctor((Functor)NonTagPart(t));
      break;
    default:
      abort_optyap("unknown type tag in function show_answer_trie");
  }

  show_subgoal_trie(TrNode_child(sg_node), str, str_index, arity);
  return;
}


void show_answer_trie(ans_node_ptr ans_node, char *str, int str_index, int *arity, int var_index) {
  int tag;
  Term t;
  int new_arity[100];

  if (ans_node == NULL) return;
  memcpy(new_arity, arity, 100);
  show_answer_trie(TrNode_next(ans_node), str, str_index, new_arity, var_index);

  if (arity[0] == 0) {
    str_index += sprintf(& str[str_index], "    VAR%d: ", var_index);
    var_index++;
  }

  t = TrNode_entry(ans_node);
  tag = t & TabTagBits;
  switch (tag) {
    case TabVarTagBits:
      if (arity[arity[0]] == -1) str[str_index - 1] = '|';
      str_index += sprintf(& str[str_index], "VAR%d", VarIndexOfTableTerm(t));
      while (arity[0] && (arity[arity[0]] == 1 || arity[arity[0]] == -1)) {
        if(arity[arity[0]] == 1) {
          str_index += sprintf(& str[str_index], ")");
          arity[0]--;
	}
        if(arity[arity[0]] == -1) {
          str_index += sprintf(& str[str_index], "]");
          arity[0]--;
	}
      }
      if (arity[0]) {
        if (arity[arity[0]] == -2) arity[arity[0]] = -1;
        else arity[arity[0]]--;
        str_index += sprintf(& str[str_index], ",");
      }
      break;
    case TabNumberTagBits:
      str_index += sprintf(& str[str_index], "%d", IntOfTerm(t));
      while (arity[0] && arity[arity[0]] == 1) {
        str_index += sprintf(& str[str_index], ")");
        arity[0]--;
      }
      if (arity[0]) {
        if (arity[arity[0]] == -2) arity[arity[0]] = -1;
        else arity[arity[0]]--;
        str_index += sprintf(& str[str_index], ",");
      }
      break;
    case TabAtomTagBits:
      if (arity[arity[0]] == -1) {
        if (strcmp("[]",  AtomName(AtomOfTerm(t))))
          fprintf(stdout, "    POTENCIAL ERROR: PAIR WITHOUT END ATOM '[]' !!!\n");
        str[str_index - 1] = ']';
        arity[0]--;
      } else {
        str_index += sprintf(& str[str_index], "%s", AtomName(AtomOfTerm(t)));
      }
      while (arity[0] && arity[arity[0]] == 1) {
        str_index += sprintf(& str[str_index], ")");
        arity[0]--;
      }
      if (arity[0]) {
        if (arity[arity[0]] == -2) arity[arity[0]] = -1;
        else arity[arity[0]]--;
        str_index += sprintf(& str[str_index], ",");
      }
      break;
    case TabPairTagBits:
      if (arity[arity[0]] == -1) {
        arity[arity[0]] = -2;
      } else {
        str_index += sprintf(& str[str_index], "[");
        arity[0]++;
        arity[arity[0]] = -2;
      }
      break;
    case TabApplTagBits:
      str_index += sprintf(& str[str_index], "%s(", AtomName(NameOfFunctor((Functor)NonTagPart(t))));
      arity[0]++;
      arity[arity[0]] = ArityOfFunctor((Functor)NonTagPart(t));
      break;
    default:
      abort_optyap("unknown type tag in function show_answer_trie");
  }

  if (! IS_ANSWER_LEAF_NODE(ans_node)) {
    show_answer_trie(TrNode_child(ans_node), str, str_index, arity, var_index);
  } else {
    str[str_index] = 0;
    fprintf(stdout, "%s\n", str);
  }
  return;
}


/*
static void show_subgoal_trie(sg_node_ptr sg_node, int missing_nodes, int level) {
  int i, tag, aux_nodes = -1;
  Term t;

  if (sg_node == NULL) return;
  if (missing_nodes == 0) {
    ans_node_ptr ans_node = SgFr_answer_trie((sg_fr_ptr)sg_node);
    if (! SgFr_completed((sg_fr_ptr)sg_node)) {
      for(i = 0; i < level; i++) fprintf(stdout, "  ");
      fprintf(stdout, "# ***** Potencial error: subgoal not completed !!! *****\n");
    }
    if (IS_ANSWER_LEAF_NODE(ans_node)) {
      for(i = 0; i < level; i++) fprintf(stdout, "  ");
      fprintf(stdout, "# yes\n");
    } else if (TrNode_child(ans_node) == NULL) {
      for(i = 0; i < level; i++) fprintf(stdout, "  ");
      fprintf(stdout, "# no\n");
    } else {
      show_answer_trie(TrNode_child(ans_node), level);
    }
    return;
  }
  for(i = 0; i < level; i++) fprintf(stdout, "  ");

  t = TrNode_entry(sg_node);
  tag = t & TabTagBits;
  switch (tag) {
    case TabVarTagBits:
      fprintf(stdout, "var  --> %d\n", VarIndexOfTableTerm(t));
      break;
    case TabNumberTagBits:
      fprintf(stdout, "int  --> %d\n", IntOfTerm(t));
      break;
    case TabAtomTagBits:
      fprintf(stdout, "atom --> %s\n", AtomName(AtomOfTerm(t)));
      break;
    case TabPairTagBits:
      fprintf(stdout, "pair --> start\n");
      aux_nodes += 2;
      break;
    case TabApplTagBits:
      fprintf(stdout, "appl --> %s/%d\n",
              AtomName(NameOfFunctor((Functor)NonTagPart(t))),
              ArityOfFunctor((Functor)NonTagPart(t)));
      aux_nodes += ArityOfFunctor((Functor)NonTagPart(t));
      break;
    default:
      abort_optyap("unknown type tag in function show_subgoal_trie");
  }

  show_subgoal_trie(TrNode_child(sg_node), missing_nodes + aux_nodes, level + 1);
  show_subgoal_trie(TrNode_next(sg_node), missing_nodes, level);
  return;
}


static void show_answer_trie(ans_node_ptr ans_node, int level) {
  int i, tag;
  Term t;

  if (ans_node == NULL) return;
  for(i = 0; i < level; i++) fprintf(stdout, "  ");
  fprintf(stdout, "# ");

  t = TrNode_entry(ans_node);
  tag = t & TabTagBits;
  switch (tag) {
    case TabVarTagBits:
      fprintf(stdout, "var  --> %d\n", VarIndexOfTableTerm(t));
      break;
    case TabNumberTagBits:
      fprintf(stdout, "int  --> %d\n", IntOfTerm(t));
      break;
    case TabAtomTagBits:
      fprintf(stdout, "atom --> %s\n", AtomName(AtomOfTerm(t)));
      break;
    case TabPairTagBits:
      fprintf(stdout, "pair --> start\n");
      break;
    case TabApplTagBits:
      fprintf(stdout, "appl --> %s/%d\n",
              AtomName(NameOfFunctor((Functor)NonTagPart(t))),
              ArityOfFunctor((Functor)NonTagPart(t)));
      break;
    default:
      abort_optyap("unknown type tag in function show_answer_trie");
  }

  if (! IS_ANSWER_LEAF_NODE(ans_node))
    show_answer_trie(TrNode_child(ans_node), level + 1);
  show_answer_trie(TrNode_next(ans_node), level);
  return;
}
*/



/* ------------------------- **
**      Local functions      **
** ------------------------- */

static sg_fr_ptr chain_trie_subgoals(sg_node_ptr sg_node, int missing_nodes, sg_fr_ptr last_sg_fr) {
  int tag;
  Term t;

  if (TrNode_next(sg_node))
    last_sg_fr = chain_trie_subgoals(TrNode_next(sg_node), missing_nodes, last_sg_fr);
  if (missing_nodes == 0) {
    sg_fr_ptr new_sg_fr;
    new_sg_fr = (sg_fr_ptr) TrNode_child(sg_node);
    SgFr_next(new_sg_fr) = last_sg_fr;
    return new_sg_fr;
  }

  missing_nodes -= 1;
  t = TrNode_entry(sg_node);
  tag = t & TabTagBits;
  switch (tag) {
    case TabVarTagBits:
    case TabNumberTagBits:
    case TabAtomTagBits:
      break;
    case TabPairTagBits:
      missing_nodes += 2;
      break;
    case TabApplTagBits:
      missing_nodes += ArityOfFunctor((Functor)NonTagPart(t));
      break;
    default:
      abort_optyap("unknown type tag in function chain_trie_subgoals");
  }

  last_sg_fr = chain_trie_subgoals(TrNode_child(sg_node), missing_nodes, last_sg_fr);
  return last_sg_fr;
}


static void free_subgoal_trie_branch(sg_node_ptr child_node) {
  sg_node_ptr parent_node;

  parent_node = TrNode_parent(child_node);
  while (parent_node) {
    if (TrNode_previous(child_node)) {
      TrNode_next(TrNode_previous(child_node)) = TrNode_next(child_node);
      if (TrNode_next(child_node))
        TrNode_previous(TrNode_next(child_node)) = TrNode_previous(child_node);
      FREE_SUBGOAL_TRIE_NODE(child_node);
      return;
    }
    if (TrNode_next(child_node)) {
      TrNode_previous(TrNode_next(child_node)) = NULL;
      TrNode_child(parent_node) = TrNode_next(child_node);
      FREE_SUBGOAL_TRIE_NODE(child_node);
      return;
    }
    FREE_SUBGOAL_TRIE_NODE(child_node);
    child_node = parent_node;
    parent_node = TrNode_parent(child_node);
  }
  TrNode_child(child_node) = NULL;
  return;
}


static void free_answer_trie_branch(ans_node_ptr node) {
  if (TrNode_child(node) && ! IS_ANSWER_LEAF_NODE(node))
    free_answer_trie_branch(TrNode_child(node));
  if (TrNode_next(node))
    free_answer_trie_branch(TrNode_next(node));
  FREE_ANSWER_TRIE_NODE(node);
  return;
}
#endif /* TABLING */
