/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		attvar.c						 *
* Last rev:								 *
* mods:									 *
* comments:	YAP support for attributed vars				 *
*									 *
*************************************************************************/
#ifdef SCCS
static char SccsId[]="%W% %G%";
#endif

#include "Yap.h"

#ifdef COROUTINING

#include "Yatom.h"
#include "Heap.h"
#include "heapgc.h"
#include "corout_utils.h"
#include "attvar.h"

STATIC_PROTO(Int  CurrentTime, (void));

static void AddToQueue(attvar_record *attv)
{
  Term t[2];
  sus_record *WGs;
  sus_record *new;

  t[0] = AbsAppl((CELL *)attv);
  t[1] = attv->Value;
  /* follow the chain */
  WGs = (sus_record *)ReadTimedVar(WokenGoals);
  new = (sus_record *)H;
  H = (CELL *)(new+1);
  new->NR = (sus_record *)(&(new->NR));
  new->SG = MkApplTerm(FunctorAttGoal, 2, t);
  new->NS = new;

  if ((Term)WGs == TermNil) {
    UpdateTimedVar(WokenGoals, (CELL)new);
    /* from now on, we have to start waking up goals */
    CreepFlag = Unsigned(LCL0);
  } else {
    /* add to the end of the current list of suspended goals */
    CELL *where_to = (CELL *)SDeref((CELL)WGs);
    Bind(where_to, (CELL)new);
  }
}

static CELL
DerefAttVar(CELL orig)
{
  register attvar_record *attv = RepAttVar(orig);

  return(Deref(attv->Done));
}

static CELL *
SDerefAttVar(CELL orig)
{
  register attvar_record *attv = RepAttVar(orig);

  return(SADerefa(&(attv->Done)));
}

static Term
CopyAttVar(Term orig, CELL ***to_visit_ptr)
{
  register attvar_record *attv = RepAttVar(orig);
  register attvar_record *newv;
  CELL **to_visit = *to_visit_ptr;
  Term tf = AbsAppl(H), ttime;
  Term time = CurrentTime();
  Int j;

  /* add a new attributed variable */
  newv = (attvar_record *)H;
  newv->fun = FunctorAttVar;
  newv->NS = UpdateTimedVar(AttsMutableList, tf);
  RESET_VARIABLE(&(newv->Value));
  RESET_VARIABLE(&(newv->Done));
  ttime = MkIntegerTerm(time);
  for (j = 0; j < NUM_OF_ATTS; j++) {
    newv->Atts[2*j] = ttime;
    to_visit[0] = attv->Atts+2*j;
    to_visit[1] = attv->Atts+2*j+1;
    to_visit[2] = newv->Atts+2*j+1;
    to_visit += 3;
  }
  *to_visit_ptr = to_visit;
  H = newv->Atts+2*j;
  return(tf);
}

static CELL
WakeAttVar(int bind_op, CELL reg1, CELL reg2)
{
  
  /* if bound to someone else, follow until we find the last one */
  attvar_record *attv = RepAttVar(reg1);

  /* next, do the operation */
  switch (bind_op) {

  case SBIND:
    /* we're done, so just recurse to the next level */
    if (!IsVarTerm(attv->Done) || !IsUnboundVar(attv->Done)) {
      return(unify(attv->Done,reg2));
    } else {
      /* Time to bind a variable  */
      if (reg2 != reg1) {
	Bind(&(attv->Value), reg2);
	AddToQueue(attv);
      }
      return(TRUE);
    }
    break;

  case SISPAIR:
    /* we found a variable which is going to be bound to something */
    if (IsVarTerm(attv->Value)) {
      if (IsVarTerm(attv->Done) && IsUnboundVar(attv->Done)) {
	AddToQueue(attv);
      }
      return(attv->Value);
    } else if (IsPairTerm(attv->Value))
      return(attv->Value);
      /* oopss: it had become something very different, should fail now! */
    else return(FALSE);
    break;

  case SISAPPL:
    /* we found a variable which is going to be bound to something */
    if (IsVarTerm(attv->Value)) {
      if (IsVarTerm(attv->Done) && IsUnboundVar(attv->Done)) {
	AddToQueue(attv);
      }
      return(attv->Value);
    } else if (IsApplTerm(attv->Value))
      return(attv->Value);
      /* oopss: it had become something very different, should fail now! */
    else return(FALSE);
    break;
  }
  return(0);
}

static void
mark_attvar(CELL *orig)
{
  register attvar_record *attv = (attvar_record *)orig;
  Int i;
  
  if (MARKED(((CELL)(attv->fun))))
    return;
  MARK(((CELL *)&(attv->fun)));
  total_marked ++;
  mark_variable(&(attv->Value));
  mark_variable(&(attv->Done));
  for (i = 0; i < NUM_OF_ATTS; i++) {
    if (!MARKED(attv->Atts[2*i])) {
      total_marked ++;
      MARK(attv->Atts+2*i);
    }
    mark_variable(attv->Atts+2*i+1);
  }
  mark_variable((CELL *)(&(attv->NS)));
}

static Int
CurrentTime(void) {
  if (B->cp_tr == TR) {
    /* we run the risk of not making non-determinate bindings before
       the end of the night */
    /* so we just init a TR cell that will not harm anyone */
    RESET_VARIABLE(TR);
    TR++;
  }
  return(B->cp_tr-(tr_fr_ptr)TrailBase);
}

static Int
PutAtt(attvar_record *attv, Int i, Term tatt) {
  Int pos = i*2;
  Int curt = CurrentTime();
  if (curt > IntegerOfTerm(attv->Atts[pos])) {
    MaBind(attv->Atts+pos, MkIntegerTerm(curt));    
    MaBind(attv->Atts+pos+1, tatt);
  } else {
    attv->Atts[pos+1] = tatt;
  }
  return(TRUE);
}

static Int
RmAtt(attvar_record *attv, Int i) {
  Int pos = i *2;
  if (!IsVarTerm(attv->Atts[pos+1])) {
    Int curt = CurrentTime();
    if (curt > IntegerOfTerm(attv->Atts[pos])) {
      MaBind(attv->Atts+pos, MkIntegerTerm(curt));    
      /* reset the variable */
      MaBind(attv->Atts+pos+1, (CELL)(attv->Atts+pos+1));
    } else {
      attv->Atts[pos+1] = (CELL)(attv->Atts+pos+1);
    }
  }
  return(TRUE);
}

static Int
BuildNewAttVar(Term t, Int i, Term tatt)
{
  /* allocate space in Heap */
  Term time = CurrentTime();
  int j;
  Term ttime, tout;

  attvar_record *attv = (attvar_record *)H;
  attv->fun = FunctorAttVar;
  RESET_VARIABLE(&(attv->Value));
  RESET_VARIABLE(&(attv->Done));
  ttime = MkIntegerTerm(time);
  for (j = 0; j < NUM_OF_ATTS; j++) {
    attv->Atts[2*j] = ttime;
    RESET_VARIABLE(attv->Atts+2*j+1);
  }
  H = attv->Atts+2*j;
  tout = AbsAttVar(attv);
  Bind((CELL *)t,tout);
  attv->NS = UpdateTimedVar(AttsMutableList, tout);
  return(PutAtt(attv, i, tatt));
}

static Int
GetAtt(attvar_record *attv, int i) {
  Int pos = i *2;
  return(attv->Atts[pos+1]);
}

static Int
FreeAtt(attvar_record *attv, int i) {
  Int pos = i *2;
  return(IsVarTerm(attv->Atts[pos+1]));
}

static Int
BindAttVar(attvar_record *attv) {
  return(unify(attv->Done, attv->Value));  
}

static Term
GetAllAtts(attvar_record *attv) {
  Int i;
  Term t = TermNil;
  for (i = NUM_OF_ATTS-1; i >= 0; i --) {
    if (!IsVarTerm(attv->Atts[2*i+1]))
      t = MkPairTerm(attv->Atts[2*i+1], t);
  }
  return(t);
}

static Term
AllAttVars(Term t) {
  if (t == TermNil) {
    return(t);
  } else {
    attvar_record *attv = RepAttVar(t);
    if (!IsVarTerm(attv->Done) || !IsUnboundVar(attv->Done))
      return(AllAttVars(attv->NS));
    else return(MkPairTerm(t,AllAttVars(attv->NS)));
  }
}

Term
CurrentAttVars(void) {
  return(AllAttVars(ReadTimedVar(AttsMutableList)));

}

static Int
p_put_att(void) {
  /* receive a variable in ARG1 */
  Term inp = SDeref(ARG1);
 restart:
  /* if this is unbound, ok */
  if (IsVarTerm(inp)) {
    return(BuildNewAttVar(inp, IntegerOfTerm(Deref(ARG2)), Deref(ARG3)));
  }
  if (IsApplTerm(inp)) {
    attvar_record *attv;
    Functor fun = FunctorOfTerm(inp);
    if (fun != FunctorAttVar) {
      Error(TYPE_ERROR_VARIABLE,inp,"put_attributes/2");
      return(FALSE);
    }
    attv = RepAttVar(inp);
    if (!IsVarTerm(attv->Done) || !IsUnboundVar(attv->Done)) {
      inp = SDeref(attv->Value);
      goto restart;
    }
    return(PutAtt(attv, IntegerOfTerm(Deref(ARG2)), Deref(ARG3)));
  } else {
    Error(TYPE_ERROR_VARIABLE,inp,"put_attributes/2");
    return(FALSE);
  }
}

static Int
p_rm_att(void) {
  /* receive a variable in ARG1 */
  Term inp = SDeref(ARG1);
 restart:
  /* if this is unbound, ok */
  if (IsVarTerm(inp)) {
    return(TRUE);
  }
  if (IsApplTerm(inp)) {
    attvar_record *attv;
    Functor fun = FunctorOfTerm(inp);
    if (fun != FunctorAttVar) {
      Error(TYPE_ERROR_VARIABLE,inp,"delete_attributes/2");
      return(FALSE);
    }
    attv = RepAttVar(inp);
    if (!IsVarTerm(attv->Done) || !IsUnboundVar(attv->Done)) {
      inp = SDeref(attv->Value);
      goto restart;
    }
    return(RmAtt(attv, IntegerOfTerm(Deref(ARG2))));
  } else {
    Error(TYPE_ERROR_VARIABLE,inp,"delete_attribute/2");
    return(FALSE);
  }
}

static Int
p_get_att(void) {
  /* receive a variable in ARG1 */
  Term  inp = SDeref(ARG1);
  /* if this is unbound, ok */
 restart:
  if (IsVarTerm(inp)) {
    /*    Error(INSTANTIATION_ERROR,inp,"get_att/2");*/
    return(FALSE);
  }
  if (IsApplTerm(inp)) {
    Functor fun = FunctorOfTerm(inp);
    attvar_record *attv;
    Term out;
    if (fun != FunctorAttVar) {
      Error(TYPE_ERROR_VARIABLE,inp,"get_att/2");
      return(FALSE);
    }
    attv = RepAttVar(inp);
    if (!IsVarTerm(attv->Done) || !IsUnboundVar(attv->Done)) {
      inp = SDeref(attv->Value);
      goto restart;
    }
    out = GetAtt(attv,IntegerOfTerm(Deref(ARG2)));
    return(!IsVarTerm(out) && unify(ARG3,out));
  } else {
    Error(TYPE_ERROR_VARIABLE,inp,"get_att/2");
    return(FALSE);
  }
}

static Int
p_free_att(void) {
  /* receive a variable in ARG1 */
  Term  inp = SDeref(ARG1);
 restart:
  /* if this is unbound, ok */
  if (IsVarTerm(inp)) {
    return(TRUE);
  }
  if (IsApplTerm(inp)) {
    Functor fun = FunctorOfTerm(inp);
    attvar_record *attv;
    if (fun != FunctorAttVar) {
      Error(TYPE_ERROR_VARIABLE,inp,"free_att/2");
      return(FALSE);
    }
    attv = RepAttVar(inp);
    if (!IsVarTerm(attv->Done) || !IsUnboundVar(attv->Done)) {
      inp = SDeref(attv->Value);
      goto restart;
    }
    return(FreeAtt(attv,IntegerOfTerm(Deref(ARG2))));
  } else {
    Error(TYPE_ERROR_VARIABLE,inp,"free_att/2");
    return(FALSE);
  }
}

static Int
p_bind_attvar(void) {
  /* receive a variable in ARG1 */
  Term  inp = SDeref(ARG1);
 restart:
  /* if this is unbound, ok */
  if (IsVarTerm(inp)) {
    return(TRUE);
  }
  if (IsApplTerm(inp)) {
    attvar_record *attv;
    Functor fun = FunctorOfTerm(inp);
    if (fun != FunctorAttVar) {
      Error(TYPE_ERROR_VARIABLE,inp,"bind_att/2");
      return(FALSE);
    }
    attv = RepAttVar(inp);
    if (!IsVarTerm(attv->Done) || !IsUnboundVar(attv->Done)) {
      inp = SDeref(attv->Value);
      goto restart;
    }
    return(BindAttVar(attv));
  } else {
    Error(TYPE_ERROR_VARIABLE,inp,"bind_att/2");
    return(FALSE);
  }
}

static Int
p_get_all_atts(void) {
  /* receive a variable in ARG1 */
  Term inp = SDeref(ARG1);
 restart:
  /* if this is unbound, ok */
  if (IsVarTerm(inp)) {
    return(TRUE);
  }
  if (IsApplTerm(inp)) {
    attvar_record *attv;
    Functor fun = FunctorOfTerm(inp);
    if (fun != FunctorAttVar) {
      Error(TYPE_ERROR_VARIABLE,inp,"get_att/2");
      return(FALSE);
    }
    attv = RepAttVar(inp);
    if (!IsVarTerm(attv->Done) || !IsUnboundVar(attv->Done)) {
      inp = SDeref(attv->Value);
      goto restart;
    }
    return(unify(ARG3,GetAllAtts(attv)));
  } else {
    Error(TYPE_ERROR_VARIABLE,inp,"get_att/2");
    return(FALSE);
  }
}

static Int
p_inc_atts(void)
{
  Term t = MkIntegerTerm(NUM_OF_ATTS);
  NUM_OF_ATTS++;
  return(unify(ARG1,t));
}

static Int
p_n_atts(void)
{
  Term t = MkIntegerTerm(NUM_OF_ATTS);
  return(unify(ARG1,t));
}

static Int
p_all_attvars(void)
{
  Term t = ReadTimedVar(AttsMutableList);
  return(unify(ARG1,AllAttVars(t)));
}

static Int
p_is_attvar(void)
{
  Term t = SDeref(ARG1);
  return(!IsVarTerm(t) &&
	 IsApplTerm(t) &&
	 FunctorOfTerm(t) == FunctorAttVar);
}

void InitAttVarPreds(void)
{
  attas[attvars_ext].bind_op = WakeAttVar;
  attas[attvars_ext].deref_op = DerefAttVar;
  attas[attvars_ext].sderef_op = SDerefAttVar;
  attas[attvars_ext].copy_term_op = CopyAttVar;
  attas[attvars_ext].binding_transparent = TRUE;
  attas[attvars_ext].mark_op = mark_attvar;
  InitCPred("get_att", 3, p_get_att, SafePredFlag);
  InitCPred("get_all_atts", 2, p_get_all_atts, SafePredFlag);
  InitCPred("free_att", 2, p_free_att, SafePredFlag);
  InitCPred("put_att", 3, p_put_att, SafePredFlag);
  InitCPred("rm_att", 2, p_rm_att, SafePredFlag);
  InitCPred("inc_n_of_atts", 1, p_inc_atts, SafePredFlag);
  InitCPred("n_of_atts", 1, p_n_atts, SafePredFlag);
  InitCPred("bind_attvar", 1, p_bind_attvar, SafePredFlag);
  InitCPred("$all_attvars", 1, p_all_attvars, SafePredFlag);
  InitCPred("$is_att_variable", 1, p_is_attvar, SafePredFlag);
}

#endif /* COROUTINING */


