/*************************************************************************
*									 *
*	 YAP Prolog   %W% %G%
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		sshift.h						 *
* Last rev:	19/2/88							 *
* mods:									 *
* comments:	stack shifter functionality for YAP			 *
*									 *
*************************************************************************/

/* use an auxiliary function for ranges */
#define IN_BETWEEN(MIN,X,MAX) (Unsigned((Int)(X)-(Int)(MIN)) <=  \
 			    Unsigned((Int)(MAX)-(Int)(MIN)) )


/* The difference between the old stack pointers and the new ones */
extern Int HDiff,
  GDiff,
  LDiff,
  TrDiff,
  XDiff;

/* The old stack pointers */
extern CELL    *OldASP, *OldLCL0;
extern tr_fr_ptr OldTR;
extern CELL    *OldH0, *OldH;
extern ADDR     OldTrailBase, OldTrailTop;
extern ADDR     OldHeapBase, OldHeapTop;

#define CharP(ptr)	((char *) (ptr))

Inline(IsHeapP, int, CELL *, ptr, (ptr >= (CELL *)HeapBase && ptr <= (CELL *)HeapTop) )

/* Adjusting cells and pointers to cells */
Inline(PtoGloAdjust, CELL *, CELL *, ptr, ((CELL *)(CharP(ptr) + GDiff)) )
Inline(PtoTRAdjust, tr_fr_ptr, tr_fr_ptr, ptr, ((tr_fr_ptr)(CharP(ptr) + TrDiff)) )
Inline(PtoLocAdjust, CELL *, CELL *, ptr, ((CELL *)(CharP(ptr) + LDiff)) )

Inline(GlobalAdjust, CELL, CELL, val, (val+GDiff) )
Inline(GlobalAddrAdjust, ADDR, ADDR, ptr, (ptr+GDiff) )
Inline(LocalAdjust, CELL, CELL, val, (val+LDiff) )
Inline(LocalAddrAdjust, ADDR, ADDR, ptr, (ptr+LDiff) )
Inline(TrailAdjust, CELL, CELL, val, (val+TrDiff) )
Inline(TrailAddrAdjust, ADDR, ADDR, ptr, (ptr+TrDiff) )
/* heap data structures */
#if	USE_OFFSETS
Inline(AtomAdjust, Atom, Atom, at, (at) )
Inline(AtomTermAdjust, Term, Term, at, (at) )
Inline(FuncAdjust, Functor, Functor, f, (f) )
Inline(PropAdjust, Prop, Prop, p, (p) )
#else
Inline(AtomAdjust, Atom, Atom, at, (Atom)(CharP(at)+HDiff) )
#if MMAP_ADDR >= 0x40000000
Inline(AtomTermAdjust, Term, Term, at, (at) )
#else
Inline(AtomTermAdjust, Term, Term, at, MkAtomTerm((Atom)(CharP(AtomOfTerm(at)+HDiff))) )
#endif
Inline(FuncAdjust, Functor, Functor, f, (Functor)(CharP(f)+HDiff) )
Inline(PropAdjust, Prop, Prop, p, (Prop)(CharP(p)+HDiff) )
#endif
Inline(ConsultObjAdjust, consult_obj *, consult_obj *, co, (consult_obj *)(CharP(co)+HDiff) )
Inline(DBRefAdjust, DBRef, DBRef, dbr, (DBRef)(CharP(dbr)+HDiff) )
Inline(CodeAdjust, Term, Term, dbr, ((Term)(dbr)+HDiff) )
Inline(AddrAdjust, ADDR, ADDR, addr, (ADDR)(CharP(addr)+HDiff) )
Inline(CodeAddrAdjust, CODEADDR, CODEADDR, addr, (CODEADDR)(CharP(addr)+HDiff) )
Inline(BlockAdjust, BlockHeader *, BlockHeader *, addr, (BlockHeader *)(CharP(addr)+HDiff) )
Inline(PtoOpAdjust, yamop *, yamop *, ptr, ((yamop *)(CharP(ptr) + HDiff)) )
Inline(PtoHeapCellAdjust, CELL *, CELL *, ptr, ((CELL *)(CharP(ptr) + HDiff)) )
#if PRECOMPUTE_REGADDRESS
Inline(XAdjust, AREG, AREG, reg, (AREG)((reg)+XDiff) )
#else
Inline(XAdjust, AREG, AREG, reg, (reg) )
#endif
Inline(YAdjust, YREG, YREG, reg, (reg) )

Inline(IsOldLocal, int, CELL, reg, IN_BETWEEN(OldASP, reg, OldLCL0))

/* require because the trail might contain dangling pointers */
Inline(IsOldLocalInTR, int, CELL, reg, IN_BETWEEN(OldH, reg, OldLCL0) )

Inline(IsOldH, int, CELL, reg, ( CharP(reg) == CharP(OldH) ) )


Inline(IsOldGlobal, int, CELL, reg, IN_BETWEEN(OldH0, reg, OldH) )
Inline(IsOldGlobalPtr, int, CELL *, ptr, IN_BETWEEN( OldH0, ptr, OldH) )
Inline(IsOldTrail, int, CELL, reg, IN_BETWEEN(OldTrailBase, reg, OldTR) )
Inline(IsOldCode, int, CELL, reg, IN_BETWEEN(OldHeapBase, reg, OldHeapTop) )
Inline(IsGlobal, int, CELL, reg, IN_BETWEEN(H0, reg, H) )

void STD_PROTO(AdjustStacksAndTrail, (void));
void STD_PROTO(AdjustRegs, (int));
